@@ -40,6 +40,7 @@
 #include <syslog.h>
 #include <sys/time.h>
 #include <errno.h>
+#include <getopt.h>
 
 /* Typical include path would be <librdkafka/rdkafka.h>, but this program
  * is builtin from within the librdkafka source tree and thus differs. */
@@ -237,7 +238,7 @@ static int describe_groups (rd_kafka_t *rk, const char *group) {
                 const struct rd_kafka_group_info *gi = &grplist->groups[i];
                 int j;
 
-                printf("Group \"%s\" in state %s on broker %d (%s:%hu)\n",
+                printf("Group \"%s\" in state %s on broker %d (%s:%d)\n",
                        gi->group, gi->state,
                        gi->broker.id, gi->broker.host, gi->broker.port);
                 if (gi->err)
