@@ -46,6 +46,7 @@ typedef int mode_t;
 namespace RdKafka {
 
 
+void consume_cb_trampoline(rd_kafka_message_t *msg, void *opaque);
 void log_cb_trampoline (const rd_kafka_t *rk, int level,
                         const char *fac, const char *buf);
 void error_cb_trampoline (rd_kafka_t *rk, int err, const char *reason,
@@ -215,7 +216,8 @@ private:
 class ConfImpl : public Conf {
  public:
   ConfImpl()
-      :dr_cb_(NULL),
+      :consume_cb_(NULL),
+      dr_cb_(NULL),
       event_cb_(NULL),
       socket_cb_(NULL),
       open_cb_(NULL),
@@ -236,6 +238,22 @@ class ConfImpl : public Conf {
                        const std::string &value,
                        std::string &errstr);
 
+  Conf::ConfResult set (const std::string &name, ConsumeCb *consume_cb,
+                        std::string &errstr) {
+    if (name != "consume_cb") {
+      errstr = "Invalid value type";
+      return Conf::CONF_INVALID;
+    }
+
+    if (!rk_conf_) {
+      errstr = "Requires RdKafka::Conf::CONF_GLOBAL object";
+      return Conf::CONF_INVALID;
+    }
+
+    consume_cb_ = consume_cb;
+    return Conf::CONF_OK;
+  }
+
   Conf::ConfResult set (const std::string &name, DeliveryReportCb *dr_cb,
                         std::string &errstr) {
     if (name != "dr_cb") {
@@ -421,6 +439,7 @@ class ConfImpl : public Conf {
 
   std::list<std::string> *dump ();
 
+  ConsumeCb *consume_cb_;
   DeliveryReportCb *dr_cb_;
   EventCb *event_cb_;
   SocketCb *socket_cb_;
@@ -483,6 +502,7 @@ class HandleImpl : virtual public Handle {
    * the opaque provided to rdkafka must be a pointer to HandleImpl, since
    * ProducerImpl and ConsumerImpl classes cannot be safely directly cast to
    * HandleImpl due to the skewed diamond inheritance. */
+  ConsumeCb *consume_cb_;
   EventCb *event_cb_;
   SocketCb *socket_cb_;
   OpenCb *open_cb_;
@@ -535,6 +555,10 @@ public:
   topic_(topic), partition_(partition), offset_(RdKafka::Topic::OFFSET_INVALID),
       err_(ERR_NO_ERROR) {}
 
+  TopicPartitionImpl (const std::string &topic, int partition, int64_t offset):
+  topic_(topic), partition_(partition), offset_(offset),
+          err_(ERR_NO_ERROR) {}
+
   TopicPartitionImpl (const rd_kafka_topic_partition_t *c_part) {
     topic_ = std::string(c_part->topic);
     partition_ = c_part->partition;
@@ -546,9 +570,9 @@ public:
   int partition () { return partition_; }
   const std::string &topic () const { return topic_ ; }
 
-  int64_t offset () { return offset_; }
+  int64_t offset () const { return offset_; }
 
-  ErrorCode err () { return err_; }
+  ErrorCode err () const { return err_; }
 
   void set_offset (int64_t offset) { offset_ = offset; }
 
@@ -677,6 +701,8 @@ class ConsumerImpl : virtual public Consumer, virtual public HandleImpl {
   ErrorCode start (Topic *topic, int32_t partition, int64_t offset,
                    Queue *queue);
   ErrorCode stop (Topic *topic, int32_t partition);
+  ErrorCode seek (Topic *topic, int32_t partition, int64_t offset,
+		  int timeout_ms);
   Message *consume (Topic *topic, int32_t partition, int timeout_ms);
   Message *consume (Queue *queue, int timeout_ms);
   int consume_callback (Topic *topic, int32_t partition, int timeout_ms,
@@ -709,6 +735,11 @@ class ProducerImpl : virtual public Producer, virtual public HandleImpl {
                      const std::vector<char> *key,
                      void *msg_opaque);
 
+  ErrorCode flush (int timeout_ms) {
+	  return static_cast<RdKafka::ErrorCode>(rd_kafka_flush(rk_,
+								timeout_ms));
+  }
+
   static Producer *create (Conf *conf, std::string &errstr);
 
 };
