@@ -204,6 +204,11 @@ static RD_INLINE RD_UNUSED void *rd_memdup (const void *src, size_t size) {
 	return dst;
 }
 
+/**
+ * @brief Memset &OBJ to 0, does automatic sizeof(OBJ).
+ */
+#define RD_MEMZERO(OBJ) memset(&(OBJ), 0, sizeof(OBJ))
+
 
 /**
  * Generic refcnt interface
@@ -229,7 +234,7 @@ static RD_INLINE RD_UNUSED int rd_refcnt_init (rd_refcnt_t *R, int v) {
         return r;
 }
 #else
-#define rd_refcnt_init(R,v)  rd_atomic32_set(R, v)
+#define rd_refcnt_init(R,v)  rd_atomic32_init(R, v)
 #endif
 
 #ifdef RD_REFCNT_USE_LOCKS
@@ -355,6 +360,8 @@ static RD_INLINE RD_UNUSED int rd_refcnt_get (rd_refcnt_t *R) {
 #define RD_SHARED_PTR_TYPE(STRUCT_NAME,WRAPPED_TYPE) WRAPPED_TYPE
 
 
+#define rd_shared_ptr_get_src(FUNC,LINE,OBJ,REFCNT,SPTR_TYPE)	\
+        (rd_refcnt_add(REFCNT), (OBJ))
 #define rd_shared_ptr_get(OBJ,REFCNT,SPTR_TYPE)          \
         (rd_refcnt_add(REFCNT), (OBJ))
 
@@ -402,10 +409,12 @@ rd_shptr0_t *rd_shared_ptr_get0 (const char *func, int line,
         return sptr;
 }
 
-#define rd_shared_ptr_get(OBJ,REF,SPTR_TYPE)                            \
+#define rd_shared_ptr_get_src(FUNC,LINE,OBJ,REF,SPTR_TYPE)		\
         (rd_refcnt_add(REF),                                            \
-         (SPTR_TYPE *)rd_shared_ptr_get0(__FUNCTION__,__LINE__,         \
-                                         #SPTR_TYPE,REF,OBJ))
+         (SPTR_TYPE *)rd_shared_ptr_get0(FUNC,LINE, #SPTR_TYPE,REF,OBJ))
+#define rd_shared_ptr_get(OBJ,REF,SPTR_TYPE)	\
+	rd_shared_ptr_get_src(__FUNCTION__, __LINE__, OBJ, REF, SPTR_TYPE)
+
 
 
 #define rd_shared_ptr_obj(SPTR) (SPTR)->obj
@@ -421,3 +430,6 @@ rd_shptr0_t *rd_shared_ptr_get0 (const char *func, int line,
 
 void rd_shared_ptrs_dump (void);
 #endif
+
+
+#define RD_IF_FREE(PTR,FUNC) do { if ((PTR)) FUNC(PTR); } while (0)
