@@ -35,6 +35,8 @@ void rd_kafka_buf_destroy_final (rd_kafka_buf_t *rkbuf) {
         if (rkbuf->rkbuf_response)
                 rd_kafka_buf_destroy(rkbuf->rkbuf_response);
 
+	rd_kafka_replyq_destroy(&rkbuf->rkbuf_replyq);
+
         if (rkbuf->rkbuf_buf2)
                 rd_free(rkbuf->rkbuf_buf2);
 
@@ -85,6 +87,7 @@ void rd_kafka_buf_rewind (rd_kafka_buf_t *rkbuf, int iovindex, size_t new_of,
 	rkbuf->rkbuf_msg.msg_iovlen = iovindex;
 	rkbuf->rkbuf_wof = new_of;
 	rkbuf->rkbuf_wof_init = new_of_init;
+	rkbuf->rkbuf_len = new_of;
 
 }
 
@@ -100,14 +103,19 @@ struct iovec *rd_kafka_buf_iov_next (rd_kafka_buf_t *rkbuf) {
  * Pushes 'buf' & 'len' onto the previously allocated iov stack for 'rkbuf'.
  */
 void rd_kafka_buf_push0 (rd_kafka_buf_t *rkbuf, const void *buf, size_t len,
-			 int allow_crc_calc) {
+			 int allow_crc_calc, int auto_push) {
 	struct iovec *iov;
 
+	if (auto_push)
+		rd_kafka_buf_autopush(rkbuf);
+
 	iov = rd_kafka_buf_iov_next(rkbuf);
 
 	iov->iov_base = (void *)buf;
 	iov->iov_len = len;
 
+	rkbuf->rkbuf_len += len;
+
 	if (allow_crc_calc && (rkbuf->rkbuf_flags & RD_KAFKA_OP_F_CRC))
 		rkbuf->rkbuf_crc = rd_crc32_update(rkbuf->rkbuf_crc, buf, len);
 }
@@ -121,10 +129,11 @@ void rd_kafka_buf_push0 (rd_kafka_buf_t *rkbuf, const void *buf, size_t len,
  *   once and after all buf_write()s have been performed.
  */
 void rd_kafka_buf_autopush (rd_kafka_buf_t *rkbuf) {
-	rd_kafka_assert(NULL, rkbuf->rkbuf_wof > rkbuf->rkbuf_wof_init);
-        rd_kafka_buf_push0(rkbuf, rkbuf->rkbuf_wbuf + rkbuf->rkbuf_wof_init,
-			   rkbuf->rkbuf_wof - rkbuf->rkbuf_wof_init,
-			   0/* No CRC calc */);
+	if (rkbuf->rkbuf_wof > rkbuf->rkbuf_wof_init)
+		rd_kafka_buf_push0(rkbuf,
+				   rkbuf->rkbuf_wbuf + rkbuf->rkbuf_wof_init,
+				   rkbuf->rkbuf_wof - rkbuf->rkbuf_wof_init,
+				   0/* No CRC calc */, 0/* No autopush */);
 }
 
 
@@ -204,7 +213,6 @@ rd_kafka_buf_t *rd_kafka_buf_new0 (const rd_kafka_t *rk,
         rkbuf->rkbuf_flags = flags;
 	rkbuf->rkbuf_iov = (struct iovec *)(rkbuf+1);
 	rkbuf->rkbuf_iovcnt = iovcnt;
-	rd_kafka_assert(NULL, rkbuf->rkbuf_iovcnt <= IOV_MAX);
 	rkbuf->rkbuf_msg.msg_iov = rkbuf->rkbuf_iov;
 
         if (growable)
@@ -216,9 +224,7 @@ rd_kafka_buf_t *rd_kafka_buf_new0 (const rd_kafka_t *rk,
 
         if (rk) {
                 /* save the first iovecs for the header + clientid */
-                rkbuf->rkbuf_iov[0].iov_base = rkbuf->rkbuf_buf;
-                rkbuf->rkbuf_iov[0].iov_len  = headersize;
-                rkbuf->rkbuf_msg.msg_iovlen = 1;
+		rd_kafka_buf_push0(rkbuf, rkbuf->rkbuf_buf, headersize, 0, 0);
         }
 
 	rkbuf->rkbuf_size     = size;
@@ -269,8 +275,8 @@ void rd_kafka_bufq_deq (rd_kafka_bufq_t *rkbufq, rd_kafka_buf_t *rkbuf) {
 
 void rd_kafka_bufq_init(rd_kafka_bufq_t *rkbufq) {
 	TAILQ_INIT(&rkbufq->rkbq_bufs);
-	rd_atomic32_set(&rkbufq->rkbq_cnt, 0);
-	rd_atomic32_set(&rkbufq->rkbq_msg_cnt, 0);
+	rd_atomic32_init(&rkbufq->rkbq_cnt, 0);
+	rd_atomic32_init(&rkbufq->rkbq_msg_cnt, 0);
 }
 
 /**
@@ -306,20 +312,23 @@ void rd_kafka_bufq_purge (rd_kafka_broker_t *rkb,
 
 
 /**
- * @brief Purge connection-setup API requests from the queue.
+ * @brief Update bufq for connection reset:
+ *
+ * - Purge connection-setup API requests from the queue.
+ * - Reset any partially sent buffer's offset. (issue #756)
  *
  * Request types purged:
  *   ApiVersion
  *   SaslHandshake
  */
-void rd_kafka_bufq_purge_connsetup (rd_kafka_broker_t *rkb,
-				    rd_kafka_bufq_t *rkbufq) {
+void rd_kafka_bufq_connection_reset (rd_kafka_broker_t *rkb,
+				     rd_kafka_bufq_t *rkbufq) {
 	rd_kafka_buf_t *rkbuf, *tmp;
 
 	rd_kafka_assert(rkb->rkb_rk, thrd_is_current(rkb->rkb_thread));
 
-	rd_rkb_dbg(rkb, QUEUE, "BUFQ", "Purging connection-setup requests "
-		   "from bufq with %i buffers",
+	rd_rkb_dbg(rkb, QUEUE, "BUFQ",
+		   "Updating %d buffers on connection reset",
 		   rd_atomic32_get(&rkbufq->rkbq_cnt));
 
 	TAILQ_FOREACH_SAFE(rkbuf, &rkbufq->rkbq_bufs, rkbuf_link, tmp) {
@@ -333,16 +342,47 @@ void rd_kafka_bufq_purge_connsetup (rd_kafka_broker_t *rkb,
 					      NULL, rkbuf);
 			break;
 		default:
+			if (rkbuf->rkbuf_of > 0)
+				rkbuf->rkbuf_of = 0;
 			break;
 		}
         }
 }
 
 
-size_t rd_kafka_buf_write_Message (rd_kafka_buf_t *rkbuf,
+void rd_kafka_bufq_dump (rd_kafka_broker_t *rkb, const char *fac,
+			 rd_kafka_bufq_t *rkbq) {
+	rd_kafka_buf_t *rkbuf;
+	int cnt = rd_kafka_bufq_cnt(rkbq);
+	rd_ts_t now;
+
+	if (!cnt)
+		return;
+
+	now = rd_clock();
+
+	rd_rkb_dbg(rkb, BROKER, fac, "bufq with %d buffer(s):", cnt);
+
+	TAILQ_FOREACH(rkbuf, &rkbq->rkbq_bufs, rkbuf_link) {
+		rd_rkb_dbg(rkb, BROKER, fac,
+			   " Buffer %s (%"PRIdsz" bytes, corrid %"PRId32", "
+			   "connid %d, retry %d in %lldms, timeout in %lldms",
+			   rd_kafka_ApiKey2str(rkbuf->rkbuf_reqhdr.ApiKey),
+			   rkbuf->rkbuf_len, rkbuf->rkbuf_corrid,
+			   rkbuf->rkbuf_connid, rkbuf->rkbuf_retries,
+			   rkbuf->rkbuf_ts_retry ?
+			   (now - rkbuf->rkbuf_ts_retry) / 1000LL : 0,
+			   rkbuf->rkbuf_ts_timeout ?
+			   (now - rkbuf->rkbuf_ts_timeout) / 1000LL : 0);
+	}
+}
+
+
+size_t rd_kafka_buf_write_Message (rd_kafka_broker_t *rkb,
+				   rd_kafka_buf_t *rkbuf,
 				   int64_t Offset, int8_t MagicByte,
 				   int8_t Attributes, int64_t Timestamp,
-				   const rd_kafkap_bytes_t *key,
+				   const void *key, int32_t key_len,
 				   const void *payload, int32_t len,
 				   int *outlenp) {
 	int32_t MessageSize;
@@ -357,8 +397,8 @@ size_t rd_kafka_buf_write_Message (rd_kafka_buf_t *rkbuf,
 
 	/* MessageSize */
 	MessageSize = (4 + 1 + 1 + /* Crc+MagicByte+Attributes */
-		       RD_KAFKAP_BYTES_SIZE(key) +
-		       4 /* Message.ValueLength */ +
+		       4 /* KeyLength */ + key_len +
+		       4 /* ValueLength */ +
 		       len /* Value length */);
 	if (MagicByte == 1)
 		MessageSize += 8; /* Timestamp i64 */
@@ -383,22 +423,27 @@ size_t rd_kafka_buf_write_Message (rd_kafka_buf_t *rkbuf,
 	/* V1: Timestamp */
 	if (MagicByte == 1)
 		rd_kafka_buf_write_i64(rkbuf, Timestamp);
-	
-	/* Push write-buffer onto iovec stack */
-        rd_kafka_buf_autopush(rkbuf);
 
 	/* Message Key */
-	rd_kafka_buf_push_kbytes(rkbuf, key);
+	if (key) {
+		rd_kafka_buf_write_i32(rkbuf, key_len);
+		rd_kafka_buf_write(rkbuf, key, key_len);
+	} else {
+		rd_kafka_buf_write_i32(rkbuf, RD_KAFKAP_BYTES_LEN_NULL);
+	}
 
 	/* Value(payload) length */
-	rd_kafka_buf_write_i32(rkbuf, payload ? len : RD_KAFKAP_BYTES_LEN_NULL);
-
-	/* Push write-buffer onto iovec stack */
-        rd_kafka_buf_autopush(rkbuf);
-
-	/* Value */
-	if (payload)
-		rd_kafka_buf_push(rkbuf, payload, len);
+	if (payload) {
+		/* If payload is below the copy limit we copy the payload
+		 * to the buffer, otherwise we push it onto the iovec stack.*/
+		rd_kafka_buf_write_i32(rkbuf, len);
+		if (len <= rkb->rkb_rk->rk_conf.msg_copy_max_size)
+			rd_kafka_buf_write(rkbuf, payload, len);
+		else
+			rd_kafka_buf_push(rkbuf, payload, len);
+	} else {
+		rd_kafka_buf_write_i32(rkbuf, RD_KAFKAP_BYTES_LEN_NULL);
+	}
 
 	/* Finalize Crc */
 	rd_kafka_buf_update_u32(rkbuf, of_Crc,
@@ -441,12 +486,16 @@ int rd_kafka_buf_retry (rd_kafka_broker_t *rkb, rd_kafka_buf_t *rkbuf) {
 void rd_kafka_buf_handle_op (rd_kafka_op_t *rko, rd_kafka_resp_err_t err) {
         rd_kafka_buf_t *request, *response;
 
-        request = rko->rko_rkbuf;
-        rko->rko_rkbuf = NULL;
+        request = rko->rko_u.xbuf.rkbuf;
+        rko->rko_u.xbuf.rkbuf = NULL;
 
-	if (request->rkbuf_replyq) { /* NULL on op_destroy() */
-		rd_kafka_q_destroy(request->rkbuf_replyq);
-		request->rkbuf_replyq = NULL;
+	/* NULL on op_destroy() */
+	if (request->rkbuf_replyq.q) {
+		int32_t version = request->rkbuf_replyq.version;
+		rd_kafka_replyq_destroy(&request->rkbuf_replyq);
+		/* Callback might need to version check so we retain the
+		 * version across the destroy call which clears it. */
+		request->rkbuf_replyq.version = version;
 	}
 
 	if (!request->rkbuf_cb) {
@@ -491,21 +540,21 @@ void rd_kafka_buf_callback (rd_kafka_t *rk,
                 return;
 	}
 
-        if (err != RD_KAFKA_RESP_ERR__DESTROY && request->rkbuf_replyq) {
+        if (err != RD_KAFKA_RESP_ERR__DESTROY && request->rkbuf_replyq.q) {
                 rd_kafka_op_t *rko = rd_kafka_op_new(RD_KAFKA_OP_RECV_BUF);
 
 		rd_kafka_assert(NULL, !request->rkbuf_response);
 		request->rkbuf_response = response;
 
                 /* Increment refcnt since rko_rkbuf will be decref:ed
-                 * if q_enq() fails and we dont want the rkbuf gone in that
+                 * if replyq_enq() fails and we dont want the rkbuf gone in that
                  * case. */
                 rd_kafka_buf_keep(request);
-                rko->rko_rkbuf = request;
+                rko->rko_u.xbuf.rkbuf = request;
 
                 rko->rko_err = err;
 
-	        rd_kafka_q_enq(request->rkbuf_replyq, rko);
+	        rd_kafka_replyq_enq(&request->rkbuf_replyq, rko, 0);
 
 		rd_kafka_buf_destroy(request); /* from keep above */
 		return;
