@@ -34,23 +34,116 @@
 typedef struct rd_kafka_broker_s rd_kafka_broker_t;
 
 #define RD_KAFKA_HEADERS_IOV_CNT   2
-#define RD_KAFKA_PAYLOAD_IOV_MAX  (IOV_MAX-RD_KAFKA_HEADERS_IOV_CNT)
 
 /* Align X (upwards) to STRIDE, which must be power of 2. */
 #define _ALIGN(X,STRIDE) (((X) + ((STRIDE) - 1)) & -(STRIDE))
 
-/* Advance/allocate used space in marshall buffer.
- * Point PTR to available space of size LEN on success. */
-#define _MSH_ALLOC(rkbuf,PTR,LEN)  do {					\
-                size_t __LEN = (LEN);                                   \
-                if (msh_of + __LEN >= msh_size)                         \
-                        rd_kafka_buf_parse_fail(rkbuf,			\
-						"Not enough room in marshall buffer: " \
-						"%"PRIdsz"+%"PRIdsz" > %"PRIdsz, \
-						msh_of, __LEN, msh_size); \
-                (PTR) = (void *)(msh_buf+msh_of);                       \
-                msh_of += _ALIGN(__LEN, 8);				\
-        } while(0)
+
+/**
+ * Temporary buffer with memory aligned writes to accommodate
+ * effective and platform safe struct writes.
+ */
+typedef struct rd_tmpabuf_s {
+	size_t size;
+	size_t of;
+	char  *buf;
+	int    failed;
+	int    assert_on_fail;
+} rd_tmpabuf_t;
+
+/**
+ * @brief Allocate new tmpabuf with \p size bytes pre-allocated.
+ */
+static RD_UNUSED void
+rd_tmpabuf_new (rd_tmpabuf_t *tab, size_t size, int assert_on_fail) {
+	tab->buf = rd_malloc(size);
+	tab->size = size;
+	tab->of = 0;
+	tab->failed = 0;
+	tab->assert_on_fail = assert_on_fail;
+}
+
+/**
+ * @brief Free memory allocated by tmpabuf
+ */
+static RD_UNUSED void
+rd_tmpabuf_destroy (rd_tmpabuf_t *tab) {
+	rd_free(tab->buf);
+}
+
+/**
+ * @returns 1 if a previous operation failed.
+ */
+static RD_UNUSED RD_INLINE int
+rd_tmpabuf_failed (rd_tmpabuf_t *tab) {
+	return tab->failed;
+}
+
+/**
+ * @brief Allocate \p size bytes for writing returning an aligned pointer
+ *        to the memory.
+ * @returns the allocated pointer (within the tmpabuf) on success or
+ *          NULL if the requested number of bytes + alignment is not available
+ *          in the tmpabuf.
+ */
+static RD_UNUSED void *
+rd_tmpabuf_alloc0 (const char *func, int line, rd_tmpabuf_t *tab, size_t size) {
+	void *ptr;
+
+	if (unlikely(tab->failed))
+		return NULL;
+
+	if (unlikely(tab->of + size > tab->size)) {
+		if (tab->assert_on_fail) {
+			fprintf(stderr,
+				"%s: %s:%d: requested size %zd + %zd > %zd\n",
+				__FUNCTION__, func, line, tab->of, size,
+				tab->size);
+			assert(!*"rd_tmpabuf_alloc: not enough size in buffer");
+		}
+		return NULL;
+	}
+
+        ptr = (void *)(tab->buf + tab->of);
+	tab->of += _ALIGN(size, 8);
+
+	return ptr;
+}
+
+#define rd_tmpabuf_alloc(tab,size) \
+	rd_tmpabuf_alloc0(__FUNCTION__,__LINE__,tab,size)
+
+/**
+ * @brief Write \p buf of \p size bytes to tmpabuf memory in an aligned fashion.
+ *
+ * @returns the allocated and written-to pointer (within the tmpabuf) on success
+ *          or NULL if the requested number of bytes + alignment is not available
+ *          in the tmpabuf.
+ */
+static RD_UNUSED void *
+rd_tmpabuf_write0 (const char *func, int line,
+		   rd_tmpabuf_t *tab, const void *buf, size_t size) {
+	void *ptr = rd_tmpabuf_alloc0(func, line, tab, size);
+
+	if (ptr)
+		memcpy(ptr, buf, size);
+
+	return ptr;
+}
+#define rd_tmpabuf_write(tab,buf,size) \
+	rd_tmpabuf_write0(__FUNCTION__, __LINE__, tab, buf, size)
+
+
+/**
+ * @brief Wrapper for rd_tmpabuf_write() that takes a nul-terminated string.
+ */
+static RD_UNUSED char *
+rd_tmpabuf_write_str0 (const char *func, int line,
+		       rd_tmpabuf_t *tab, const char *str) {
+	return rd_tmpabuf_write0(func, line, tab, str, strlen(str)+1);
+}
+#define rd_tmpabuf_write_str(tab,str) \
+	rd_tmpabuf_write_str0(__FUNCTION__, __LINE__, tab, str)
 
 
 
@@ -66,8 +159,10 @@ typedef struct rd_kafka_broker_s rd_kafka_broker_t;
 
 #define rd_kafka_buf_parse_fail(rkbuf,...) do {				\
                 if (log_decode_errors) {                                \
+			rd_kafka_assert(NULL, rkbuf->rkbuf_rkb);	\
                         rd_rkb_log(rkbuf->rkbuf_rkb, LOG_WARNING, "PROTOERR", \
-                                   "Protocol parse failure at %s:%i",   \
+                                   "Protocol parse failure at %s:%i " \
+				   "(incorrect broker.version.fallback?)",   \
                                    __FUNCTION__, __LINE__);             \
                         rd_rkb_log(rkbuf->rkbuf_rkb, LOG_WARNING,	\
 				   "PROTOERR", __VA_ARGS__);		\
@@ -162,16 +257,24 @@ typedef struct rd_kafka_broker_s rd_kafka_broker_t;
 		rd_kafka_buf_skip(rkbuf, _ksize);			\
 	} while (0)
 
-/* Read Kafka String representation (2+N) into nul-terminated C string.
- * Depends on a marshalling environment. */
-#define rd_kafka_buf_read_str_msh(rkbuf, dst) do {			\
+/* Read Kafka String representation (2+N) and write it to the \p tmpabuf
+ * with a trailing nul byte. */
+#define rd_kafka_buf_read_str_tmpabuf(rkbuf, tmpabuf, dst) do {		\
                 rd_kafkap_str_t _kstr;					\
-		int _slen;						\
+		size_t _slen;						\
+		char *_dst;						\
 		rd_kafka_buf_read_str(rkbuf, &_kstr);			\
 		_slen = RD_KAFKAP_STR_LEN(&_kstr);			\
-                _MSH_ALLOC(rkbuf, dst, _slen+1);			\
-                memcpy(dst, _kstr.str, _slen);				\
-                dst[_slen] = '\0';					\
+		if (!(_dst =						\
+		      rd_tmpabuf_write(tmpabuf, _kstr.str, _slen+1)))	\
+			rd_kafka_buf_parse_fail(			\
+				rkbuf,					\
+				"Not enough room in tmpabuf: "		\
+				"%"PRIdsz"+%"PRIdsz			\
+				" > %"PRIdsz,				\
+				(tmpabuf)->of, _slen+1, (tmpabuf)->size); \
+		_dst[_slen] = '\0';					\
+		dst = (void *)_dst;					\
 	} while (0)
 
 /**
@@ -226,6 +329,8 @@ struct rd_kafka_buf_s { /* rd_kafka_buf_t */
 	struct msghdr rkbuf_msg;
 	struct iovec *rkbuf_iov;
 	int           rkbuf_iovcnt;
+	int     rkbuf_connid;      /* broker connection id (used when buffer
+				    * was partially sent). */
 	size_t  rkbuf_of;          /* send: send offset,
 				    * recv: parse offset */
 	size_t  rkbuf_len;         /* send: total length,
@@ -254,7 +359,7 @@ struct rd_kafka_buf_s { /* rd_kafka_buf_t */
 
 	int32_t rkbuf_expected_size;  /* expected size of message */
 
-        rd_kafka_q_t       *rkbuf_replyq;       /* Enqueue response on replyq */
+        rd_kafka_replyq_t   rkbuf_replyq;       /* Enqueue response on replyq */
         rd_kafka_resp_cb_t *rkbuf_cb;           /* Response callback */
         struct rd_kafka_buf_s *rkbuf_response;  /* Response buffer */
 
@@ -274,7 +379,6 @@ struct rd_kafka_buf_s { /* rd_kafka_buf_t */
 
         int64_t rkbuf_offset;     /* Used by OffsetCommit */
 
-	int32_t rkbuf_op_version;      /* Single Op version */
 	rd_list_t *rkbuf_rktp_vers;    /* Toppar + Op Version map.
 					* Used by FetchRequest. */
 
@@ -303,9 +407,9 @@ void rd_kafka_buf_rewind(rd_kafka_buf_t *rkbuf, int iovindex, size_t new_of,
 	size_t new_of_init);
 struct iovec *rd_kafka_buf_iov_next (rd_kafka_buf_t *rkbuf);
 void rd_kafka_buf_push0 (rd_kafka_buf_t *rkbuf, const void *buf, size_t len,
-			int allow_crc_calc);
+			 int allow_crc_calc, int auto_push);
 #define rd_kafka_buf_push(rkbuf,buf,len) \
-	rd_kafka_buf_push0(rkbuf,buf,len,1/*allow_crc*/)
+	rd_kafka_buf_push0(rkbuf,buf,len,1/*allow_crc*/, 1)
 void rd_kafka_buf_autopush (rd_kafka_buf_t *rkbuf);
 rd_kafka_buf_t *rd_kafka_buf_new_growable (const rd_kafka_t *rk,
                                            int iovcnt, size_t init_size);
@@ -320,8 +424,10 @@ void rd_kafka_bufq_concat (rd_kafka_bufq_t *dst, rd_kafka_bufq_t *src);
 void rd_kafka_bufq_purge (rd_kafka_broker_t *rkb,
                           rd_kafka_bufq_t *rkbufq,
                           rd_kafka_resp_err_t err);
-void rd_kafka_bufq_purge_connsetup (rd_kafka_broker_t *rkb,
-				    rd_kafka_bufq_t *rkbufq);
+void rd_kafka_bufq_connection_reset (rd_kafka_broker_t *rkb,
+				     rd_kafka_bufq_t *rkbufq);
+void rd_kafka_bufq_dump (rd_kafka_broker_t *rkb, const char *fac,
+			 rd_kafka_bufq_t *rkbq);
 
 int rd_kafka_buf_retry (rd_kafka_broker_t *rkb, rd_kafka_buf_t *rkbuf);
 
@@ -358,6 +464,14 @@ static RD_INLINE RD_UNUSED void rd_kafka_buf_write_seek (rd_kafka_buf_t *rkbuf,
 	rkbuf->rkbuf_wof = of;
 }
 
+/**
+ * @returns the number of bytes remaining in the write buffer.
+ */
+static RD_INLINE size_t rd_kafka_buf_write_remain (rd_kafka_buf_t *rkbuf) {
+	return rkbuf->rkbuf_size - rkbuf->rkbuf_wof;
+}
+
+
 /**
  * Write (copy) data to buffer at current write-buffer position.
  * There must be enough space allocated in the rkbuf.
@@ -580,10 +694,11 @@ static RD_INLINE size_t rd_kafka_buf_write_bytes (rd_kafka_buf_t *rkbuf,
  *
  * Returns the buffer offset of the first byte.
  */
-size_t rd_kafka_buf_write_Message (rd_kafka_buf_t *rkbuf,
+size_t rd_kafka_buf_write_Message (rd_kafka_broker_t *rkb,
+				   rd_kafka_buf_t *rkbuf,
 				   int64_t Offset, int8_t MagicByte,
 				   int8_t Attributes, int64_t Timestamp,
-				   const rd_kafkap_bytes_t *key,
+				   const void *key, int32_t key_len,
 				   const void *payload, int32_t len,
 				   int *outlenp);
 
