@@ -44,7 +44,7 @@
  */
 
 
-
+extern const char *rd_kafka_cgrp_join_state_names[];
 
 /**
  * Client group
@@ -96,12 +96,18 @@ typedef struct rd_kafka_cgrp_s {
                 /* all: waiting for previous assignment to decommission */
                 RD_KAFKA_CGRP_JOIN_STATE_WAIT_UNASSIGN,
 
-                /* all: waiting for application's rebalance_cb to operate */
-                RD_KAFKA_CGRP_JOIN_STATE_WAIT_REBALANCE_CB,
+                /* all: waiting for application's rebalance_cb to assign() */
+                RD_KAFKA_CGRP_JOIN_STATE_WAIT_ASSIGN_REBALANCE_CB,
 
-                /* all: synchronized, assigned and operational.
+		/* all: waiting for application's rebalance_cb to revoke */
+                RD_KAFKA_CGRP_JOIN_STATE_WAIT_REVOKE_REBALANCE_CB,
+
+                /* all: synchronized and assigned
                  *      may be an empty assignment. */
                 RD_KAFKA_CGRP_JOIN_STATE_ASSIGNED,
+
+		/* all: fetchers are started and operational */
+		RD_KAFKA_CGRP_JOIN_STATE_STARTED
         } rkcg_join_state;
 
         /* State when group leader */
@@ -111,9 +117,14 @@ typedef struct rd_kafka_cgrp_s {
                 int member_cnt;
         } rkcg_group_leader;
 
-        rd_kafka_q_t       rkcg_q;                  /* Application poll queue */
-        rd_kafka_q_t       rkcg_ops;                /* Manager ops queue */
-	rd_kafka_q_t       rkcg_wait_coord_q;       /* Ops awaiting coord */
+        rd_kafka_q_t      *rkcg_q;                  /* Application poll queue */
+        rd_kafka_q_t      *rkcg_ops;                /* Manager ops queue */
+	rd_kafka_q_t      *rkcg_wait_coord_q;       /* Ops awaiting coord */
+	int32_t            rkcg_version;            /* Ops queue version barrier
+						     * Increased by:
+						     *  Rebalance delegation
+						     *  Assign/Unassign
+						     */
         mtx_t              rkcg_lock;
 
         int                rkcg_flags;
@@ -126,10 +137,14 @@ typedef struct rd_kafka_cgrp_s {
                                                      *   subscription
                                                      * else:
                                                      *   static assignment */
+#define RD_KAFKA_CGRP_F_HEARTBEAT_IN_TRANSIT  0x20  /* A Heartbeat request
+                                                     * is in transit, dont
+                                                     * send a new one. */
 
         rd_interval_t      rkcg_coord_query_intvl;  /* Coordinator query intvl*/
         rd_interval_t      rkcg_heartbeat_intvl;    /* Heartbeat intvl */
         rd_interval_t      rkcg_join_intvl;         /* JoinGroup interval */
+        rd_interval_t      rkcg_timeout_scan_intvl; /* Timeout scanner */
 
         TAILQ_HEAD(, rd_kafka_topic_s)  rkcg_topics;/* Topics subscribed to */
 
@@ -158,6 +173,9 @@ typedef struct rd_kafka_cgrp_s {
 
         /* Current subscription */
         rd_kafka_topic_partition_list_t *rkcg_subscription;
+	/* The actual topics subscribed (after metadata+wildcard matching) */
+	rd_list_t *rkcg_subscribed_topics; /**< (rd_kafka_topic_info_t *) */
+
         /* Current assignment */
         rd_kafka_topic_partition_list_t *rkcg_assignment;
 
@@ -187,6 +205,18 @@ typedef struct rd_kafka_cgrp_s {
 	rd_ts_t            rkcg_ts_terminate;       /* Timestamp of when
 						     * cgrp termination was
 						     * initiated. */
+
+        /* Protected by rd_kafka_*lock() */
+        struct {
+                rd_ts_t            ts_rebalance;       /* Timestamp of
+                                                        * last rebalance */
+                int                rebalance_cnt;      /* Number of
+                                                          rebalances */
+                int                assignment_size;    /* Partition count
+                                                        * of last rebalance
+                                                        * assignment */
+        } rkcg_c;
+
 } rd_kafka_cgrp_t;
 
 
@@ -209,10 +239,10 @@ rd_kafka_cgrp_t *rd_kafka_cgrp_new (rd_kafka_t *rk,
 void rd_kafka_cgrp_serve (rd_kafka_cgrp_t *rkcg);
 
 void rd_kafka_cgrp_op (rd_kafka_cgrp_t *rkcg, rd_kafka_toppar_t *rktp,
-                       rd_kafka_q_t *replyq, rd_kafka_op_type_t type,
+                       rd_kafka_replyq_t replyq, rd_kafka_op_type_t type,
                        rd_kafka_resp_err_t err);
 void rd_kafka_cgrp_terminate0 (rd_kafka_cgrp_t *rkcg, rd_kafka_op_t *rko);
-void rd_kafka_cgrp_terminate (rd_kafka_cgrp_t *rkcg, rd_kafka_q_t *replyq);
+void rd_kafka_cgrp_terminate (rd_kafka_cgrp_t *rkcg, rd_kafka_replyq_t replyq);
 
 
 rd_kafka_resp_err_t rd_kafka_cgrp_topic_pattern_del (rd_kafka_cgrp_t *rkcg,
@@ -233,6 +263,7 @@ void rd_kafka_cgrp_handle_Metadata (rd_kafka_cgrp_t *rkcg,
                                     rd_kafka_resp_err_t err,
                                     rd_kafka_metadata_t *md);
 void rd_kafka_cgrp_handle_SyncGroup (rd_kafka_cgrp_t *rkcg,
+				     rd_kafka_broker_t *rkb,
                                      rd_kafka_resp_err_t err,
                                      const rd_kafkap_bytes_t *member_state);
 void rd_kafka_cgrp_set_join_state (rd_kafka_cgrp_t *rkcg, int join_state);
@@ -241,5 +272,6 @@ int rd_kafka_cgrp_reassign_broker (rd_kafka_cgrp_t *rkcg);
 
 void rd_kafka_cgrp_coord_query (rd_kafka_cgrp_t *rkcg,
 				const char *reason);
-
+void rd_kafka_cgrp_metadata_update_check (rd_kafka_cgrp_t *rkcg,
+					  const struct rd_kafka_metadata *md);
 #define rd_kafka_cgrp_get(rk) ((rk)->rk_cgrp)
