@@ -89,7 +89,7 @@ rd_kafka_conf_validate_broker_version (const struct rd_kafka_property *prop,
 /**
  * @brief Validate that string is a single item, without delimters (, space).
  */
-static int
+static RD_UNUSED int
 rd_kafka_conf_validate_single (const struct rd_kafka_property *prop,
 				const char *val, int ival) {
 	return !strchr(val, ',') && !strchr(val, ' ');
@@ -119,8 +119,9 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 #if WITH_SASL
 		{ 0x8, "sasl" },
 #endif
-#if HAVE_REGEX
 		{ 0x10, "regex" },
+#if WITH_LZ4
+		{ 0x20, "lz4" },
 #endif
 		{ 0, NULL }
 		}
@@ -138,6 +139,12 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	{ _RK_GLOBAL, "message.max.bytes", _RK_C_INT, _RK(max_msg_size),
 	  "Maximum transmit message size.",
 	  1000, 1000000000, 1000000 },
+	{ _RK_GLOBAL, "message.copy.max.bytes", _RK_C_INT,
+	  _RK(msg_copy_max_size),
+	  "Maximum size for message to be copied to buffer. "
+	  "Messages larger than this will be passed by reference (zero-copy) "
+	  "at the expense of larger iovecs.",
+	  0, 1000000000, 0xffff },
 	{ _RK_GLOBAL, "receive.message.max.bytes", _RK_C_INT,
           _RK(recv_max_msg_size),
 	  "Maximum receive message size. "
@@ -151,6 +158,8 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  "Maximum number of in-flight requests the client will send. "
 	  "This setting applies per broker connection.",
 	  1, 1000000, 1000000 },
+        { _RK_GLOBAL, "max.in.flight", _RK_C_ALIAS,
+          .sdef = "max.in.flight.requests.per.connection" },
 	{ _RK_GLOBAL, "metadata.request.timeout.ms", _RK_C_INT,
 	  _RK(metadata_request_timeout_ms),
 	  "Non-topic request timeout in milliseconds. "
@@ -162,6 +171,8 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  "The metadata is automatically refreshed on error and connect. "
 	  "Use -1 to disable the intervalled refresh.",
 	  -1, 3600*1000, 5*60*1000 },
+	{ _RK_GLOBAL, "metadata.max.age.ms", _RK_C_ALIAS,
+	  .sdef = "topic.metadata.refresh.interval.ms" },
 	{ _RK_GLOBAL, "topic.metadata.refresh.fast.cnt", _RK_C_INT,
 	  _RK(metadata_refresh_fast_cnt),
 	  "When a topic looses its leader this number of metadata requests "
@@ -183,19 +194,20 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
           "for matching topic names that should be ignored in "
           "broker metadata information as if the topics did not exist." },
 	{ _RK_GLOBAL, "debug", _RK_C_S2F, _RK(debug),
-	  "A comma-separated list of debug contexts to enable.",
+	  "A comma-separated list of debug contexts to enable. "
+	  "Debugging the Producer: broker,topic,msg. Consumer: cgrp,topic,fetch",
 	  .s2i = {
                         { RD_KAFKA_DBG_GENERIC,  "generic" },
 			{ RD_KAFKA_DBG_BROKER,   "broker" },
 			{ RD_KAFKA_DBG_TOPIC,    "topic" },
 			{ RD_KAFKA_DBG_METADATA, "metadata" },
-			{ RD_KAFKA_DBG_PRODUCER, "producer" },
 			{ RD_KAFKA_DBG_QUEUE,    "queue" },
 			{ RD_KAFKA_DBG_MSG,      "msg" },
 			{ RD_KAFKA_DBG_PROTOCOL, "protocol" },
                         { RD_KAFKA_DBG_CGRP,     "cgrp" },
 			{ RD_KAFKA_DBG_SECURITY, "security" },
 			{ RD_KAFKA_DBG_FETCH,    "fetch" },
+			{ RD_KAFKA_DBG_FEATURE,  "feature" },
 			{ RD_KAFKA_DBG_ALL,      "all" },
 		} },
 	{ _RK_GLOBAL, "socket.timeout.ms", _RK_C_INT, _RK(socket_timeout_ms),
@@ -219,6 +231,10 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  _RK(socket_keepalive),
           "Enable TCP keep-alives (SO_KEEPALIVE) on broker sockets",
           0, 1, 0 },
+	{ _RK_GLOBAL, "socket.nagle.disable", _RK_C_BOOL,
+	  _RK(socket_nagle_disable),
+          "Disable the Nagle algorithm (TCP_NODELAY).",
+          0, 1, 0 },
         { _RK_GLOBAL, "socket.max.fails", _RK_C_INT,
           _RK(socket_max_fails),
           "Disconnect from broker when this number of send failures "
@@ -249,6 +265,10 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  "register a stats callback using `rd_kafka_conf_set_stats_cb()`. "
 	  "The granularity is 1000ms. A value of 0 disables statistics.",
 	  0, 86400*1000, 0 },
+	{ _RK_GLOBAL, "enabled_events", _RK_C_INT,
+	  _RK(enabled_events),
+	  "See `rd_kafka_conf_set_events()`",
+	  0, 0x7fffffff, 0 },
 	{ _RK_GLOBAL, "error_cb", _RK_C_PTR,
 	  _RK(error_cb),
 	  "Error callback (set with rd_kafka_conf_set_error_cb())" },
@@ -270,7 +290,7 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  _RK(log_thread_name),
 	  "Print internal thread name in log messages "
 	  "(useful for debugging librdkafka internals)",
-	  0, 1, 0 },
+	  0, 1, ENABLE_DEVEL },
 	{ _RK_GLOBAL, "log.connection.close", _RK_C_BOOL,
 	  _RK(log_connection_close),
 	  "Log broker disconnects. "
@@ -287,6 +307,14 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
           rd_kafka_socket_cb_generic
 #endif
         },
+        { _RK_GLOBAL, "connect_cb", _RK_C_PTR,
+          _RK(connect_cb),
+          "Socket connect callback",
+        },
+        { _RK_GLOBAL, "closesocket_cb", _RK_C_PTR,
+          _RK(closesocket_cb),
+          "Socket close callback",
+        },
         { _RK_GLOBAL, "open_cb", _RK_C_PTR,
           _RK(open_cb),
           "File open callback to provide race-free CLOEXEC",
@@ -401,10 +429,6 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	 "Supported: GSSAPI, PLAIN. "
 	 "**NOTE**: Despite the name only one mechanism must be configured.",
 	 .sdef = "GSSAPI",
-	 .s2i = {
-			{ 0, "GSSAPI" },
-			{ 0, "PLAIN" }
-		},
 	 .validate = rd_kafka_conf_validate_single },
 	{ _RK_GLOBAL, "sasl.kerberos.service.name", _RK_C_STR,
 	  _RK(sasl.service_name),
@@ -502,8 +526,11 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  0, 300*1000, 100 },
         { _RK_GLOBAL|_RK_CONSUMER, "fetch.message.max.bytes", _RK_C_INT,
           _RK(fetch_msg_max_bytes),
-          "Maximum number of bytes per topic+partition to request when "
-          "fetching messages from the broker.",
+          "Initial maximum number of bytes per topic+partition to request when "
+          "fetching messages from the broker. "
+	  "If the client encounters a message larger than this value "
+	  "it will gradually try to increase it until the "
+	  "entire message can be fetched.",
           1, 1000000000, 1024*1024 },
 	{ _RK_GLOBAL|_RK_CONSUMER, "max.partition.fetch.bytes", _RK_C_ALIAS,
 	  .sdef = "fetch.message.max.bytes" },
@@ -542,13 +569,21 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  _RK(offset_commit_cb),
 	  "Offset commit result propagation callback. "
           "(set with rd_kafka_conf_set_offset_commit_cb())" },
-
+	{ _RK_GLOBAL|_RK_CONSUMER, "enable.partition.eof", _RK_C_BOOL,
+	  _RK(enable_partition_eof),
+	  "Emit RD_KAFKA_RESP_ERR__PARTITION_EOF event whenever the "
+	  "consumer reaches the end of a partition.",
+	  0, 1, 1 },
 
 	/* Global producer properties */
 	{ _RK_GLOBAL|_RK_PRODUCER, "queue.buffering.max.messages", _RK_C_INT,
 	  _RK(queue_buffering_max_msgs),
 	  "Maximum number of messages allowed on the producer queue.",
 	  1, 10000000, 100000 },
+	{ _RK_GLOBAL|_RK_PRODUCER, "queue.buffering.max.kbytes", _RK_C_INT,
+	  _RK(queue_buffering_max_kbytes),
+	  "Maximum total message size sum allowed on the producer queue.",
+	  1, INT_MAX/1024, 4000000 },
 	{ _RK_GLOBAL|_RK_PRODUCER, "queue.buffering.max.ms", _RK_C_INT,
 	  _RK(buffering_max_ms),
 	  "Maximum time, in milliseconds, for buffering data "
@@ -567,8 +602,9 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  1, 300*1000, 100 },
 	{ _RK_GLOBAL|_RK_PRODUCER, "compression.codec", _RK_C_S2I,
 	  _RK(compression_codec),
-	  "Compression codec to use for compressing message sets: "
-	  "none, gzip or snappy",
+	  "compression codec to use for compressing message sets. "
+	  "This is the default value for all topics, may be overriden by "
+	  "the topic configuration property `compression.codec`. ",
 	  .vdef = RD_KAFKA_COMPRESSION_NONE,
 	  .s2i = {
 			{ RD_KAFKA_COMPRESSION_NONE,   "none" },
@@ -577,13 +613,17 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 #endif
 #if WITH_SNAPPY
 			{ RD_KAFKA_COMPRESSION_SNAPPY, "snappy" },
+#endif
+#if WITH_LZ4
+                        { RD_KAFKA_COMPRESSION_LZ4, "lz4" },
 #endif
 			{ 0 }
 		} },
 	{ _RK_GLOBAL|_RK_PRODUCER, "batch.num.messages", _RK_C_INT,
 	  _RK(batch_num_messages),
-	  "Maximum number of messages batched in one MessageSet.",
-	  1, 1000000, 1000 },
+	  "Maximum number of messages batched in one MessageSet. "
+	  "The total MessageSet size is also limited by message.max.bytes.",
+	  1, 1000000, 10000 },
 	{ _RK_GLOBAL|_RK_PRODUCER, "delivery.report.only.error", _RK_C_BOOL,
 	  _RK(dr_err_only),
 	  "Only provide delivery reports for failed messages.",
@@ -622,7 +662,7 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  _RKT(request_timeout_ms),
 	  "The ack timeout of the producer request in milliseconds. "
 	  "This value is only enforced by the broker and relies "
-	  "on `request.required.acks` being > 0.",
+	  "on `request.required.acks` being != 0.",
 	  1, 900*1000, 5*1000 },
 	{ _RK_TOPIC|_RK_PRODUCER, "message.timeout.ms", _RK_C_INT,
 	  _RKT(message_timeout_ms),
@@ -646,8 +686,7 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  "Application opaque (set with rd_kafka_topic_conf_set_opaque())" },
 	{ _RK_TOPIC | _RK_PRODUCER, "compression.codec", _RK_C_S2I,
 	  _RKT(compression_codec),
-	  "Compression codec to use for compressing message sets: "
-	  "none, gzip or snappy",
+	  "Compression codec to use for compressing message sets. ",
 	  .vdef = RD_KAFKA_COMPRESSION_INHERIT,
 	  .s2i = {
 		  { RD_KAFKA_COMPRESSION_NONE, "none" },
@@ -657,6 +696,10 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 #if WITH_SNAPPY
 		  { RD_KAFKA_COMPRESSION_SNAPPY, "snappy" },
 #endif
+#if WITH_LZ4
+		  { RD_KAFKA_COMPRESSION_LZ4, "lz4" },
+#endif
+
 		  { RD_KAFKA_COMPRESSION_INHERIT, "inherit" },
 		  { 0 }
 		} },
@@ -672,7 +715,7 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  "`rd_kafka_offset_store()` to store an offset (optional). "
           "**NOTE:** This property should only be used with the simple "
           "legacy consumer, when using the high-level KafkaConsumer the global "
-          "`auto.commit.enable` property must be used instead. "
+          "`enable.auto.commit` property must be used instead. "
 	  "**NOTE:** There is currently no zookeeper integration, offsets "
 	  "will be written to broker or local file according to "
           "offset.store.method.",
@@ -696,8 +739,10 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  .s2i = {
 			{ RD_KAFKA_OFFSET_BEGINNING, "smallest" },
 			{ RD_KAFKA_OFFSET_BEGINNING, "earliest" },
+			{ RD_KAFKA_OFFSET_BEGINNING, "beginning" },
 			{ RD_KAFKA_OFFSET_END, "largest" },
 			{ RD_KAFKA_OFFSET_END, "latest" },
+			{ RD_KAFKA_OFFSET_END, "end" },
 			{ RD_KAFKA_OFFSET_INVALID, "error" },
 		}
 	},
@@ -762,7 +807,7 @@ rd_kafka_anyconf_set_prop0 (int scope, void *conf,
 		if (istr)
 			*str = rd_strdup(istr);
 		else
-			*str = NULL;
+			*str = prop->sdef ? rd_strdup(prop->sdef) : NULL;
 		return RD_KAFKA_CONF_OK;
 	}
         case _RK_C_KSTR:
@@ -774,7 +819,8 @@ rd_kafka_anyconf_set_prop0 (int scope, void *conf,
                 if (istr)
                         *kstr = rd_kafkap_str_new(istr, -1);
                 else
-                        *kstr = NULL;
+                        *kstr = prop->sdef ?
+				rd_kafkap_str_new(prop->sdef, -1) : NULL;
                 return RD_KAFKA_CONF_OK;
         }
 	case _RK_C_PTR:
@@ -811,15 +857,22 @@ rd_kafka_anyconf_set_prop0 (int scope, void *conf,
         {
                 /* Split comma-separated list into individual regex expressions
                  * that are verified and then append to the provided list. */
-                rd_kafka_pattern_list_t *plist;
+                rd_kafka_pattern_list_t **plist;
 
-                plist = _RK_PTR(rd_kafka_pattern_list_t *, conf, prop->offset);
+                plist = _RK_PTR(rd_kafka_pattern_list_t **, conf, prop->offset);
 
-                rd_kafka_pattern_list_clear(plist);
+		if (*plist)
+			rd_kafka_pattern_list_destroy(*plist);
+
+		if (istr) {
+			if (!(*plist =
+			      rd_kafka_pattern_list_new(istr,
+							errstr,
+                                                        (int)errstr_size)))
+				return RD_KAFKA_CONF_INVALID;
+		} else
+			*plist = NULL;
 
-                if (rd_kafka_pattern_list_init(plist, istr,
-                                               errstr, errstr_size) == -1)
-                        return RD_KAFKA_CONF_INVALID;
                 return RD_KAFKA_CONF_OK;
         }
 
@@ -882,7 +935,9 @@ rd_kafka_anyconf_set_prop (int scope, void *conf,
 		if (prop->validate &&
 		    (!value || !prop->validate(prop, value, -1))) {
 			rd_snprintf(errstr, errstr_size,
-				    "Invalid value: %s", value);
+				    "Invalid value for "
+				    "configuration property \"%s\": %s",
+				    prop->name, value);
 			return RD_KAFKA_CONF_INVALID;
 		}
 
@@ -1195,9 +1250,12 @@ static void rd_kafka_anyconf_clear (void *conf,
 
         case _RK_C_PATLIST:
         {
-                rd_kafka_pattern_list_t *plist;
-                plist = _RK_PTR(rd_kafka_pattern_list_t *, conf, prop->offset);
-                rd_kafka_pattern_list_clear(plist);
+                rd_kafka_pattern_list_t **plist;
+                plist = _RK_PTR(rd_kafka_pattern_list_t **, conf, prop->offset);
+		if (*plist) {
+			rd_kafka_pattern_list_destroy(*plist);
+			*plist = NULL;
+		}
         }
         break;
 
@@ -1287,10 +1345,11 @@ static void rd_kafka_anyconf_copy (int scope, void *dst, const void *src) {
 			break;
                 case _RK_C_PATLIST:
                 {
-                        const rd_kafka_pattern_list_t *plist;
-                        plist = _RK_PTR(const rd_kafka_pattern_list_t *,
+                        const rd_kafka_pattern_list_t **plist;
+                        plist = _RK_PTR(const rd_kafka_pattern_list_t **,
                                         src, prop->offset);
-                        val = plist->rkpl_orig;
+			if (*plist)
+				val = (*plist)->rkpl_orig;
                         break;
                 }
 		default:
@@ -1322,6 +1381,11 @@ rd_kafka_topic_conf_t *rd_kafka_topic_conf_dup (const rd_kafka_topic_conf_t
 }
 
 
+void rd_kafka_conf_set_events (rd_kafka_conf_t *conf, int events) {
+	conf->enabled_events = events;
+}
+
+
 void rd_kafka_conf_set_dr_cb (rd_kafka_conf_t *conf,
 			      void (*dr_cb) (rd_kafka_t *rk,
 					     void *payload, size_t len,
@@ -1408,6 +1472,24 @@ void rd_kafka_conf_set_socket_cb (rd_kafka_conf_t *conf,
         conf->socket_cb = socket_cb;
 }
 
+void
+rd_kafka_conf_set_connect_cb (rd_kafka_conf_t *conf,
+                              int (*connect_cb) (int sockfd,
+                                                 const struct sockaddr *addr,
+                                                 int addrlen,
+                                                 const char *id,
+                                                 void *opaque)) {
+        conf->connect_cb = connect_cb;
+}
+
+void
+rd_kafka_conf_set_closesocket_cb (rd_kafka_conf_t *conf,
+                                  int (*closesocket_cb) (int sockfd,
+                                                         void *opaque)) {
+        conf->closesocket_cb = closesocket_cb;
+}
+
+
 
 #ifndef _MSC_VER
 void rd_kafka_conf_set_open_cb (rd_kafka_conf_t *conf,
@@ -1574,10 +1656,11 @@ rd_kafka_anyconf_get0 (const void *conf, const struct rd_kafka_property *prop,
 
         case _RK_C_PATLIST:
         {
-                const rd_kafka_pattern_list_t *plist;
-                plist = _RK_PTR(const rd_kafka_pattern_list_t *,
+                const rd_kafka_pattern_list_t **plist;
+                plist = _RK_PTR(const rd_kafka_pattern_list_t **,
                                 conf, prop->offset);
-                val = plist->rkpl_orig;
+		if (*plist)
+			val = (*plist)->rkpl_orig;
                 break;
         }
 
