@@ -40,6 +40,7 @@ static const char *rd_kafka_feature_names[] = {
 	"Sasl",
 	"SaslHandshake",
 	"BrokerGroupCoordinator",
+	"LZ4",
 	NULL
 };
 
@@ -115,28 +116,44 @@ static const struct rd_kafka_feature_map {
 			{ -1 },
 		},
 
-	},
-	{
-		/* @brief >=0.9.0: SASL (GSSAPI) authentication.
-		 * Since SASL is not using the Kafka protocol
-		 * we must use something else to map us to the
-		 * proper broker version support:
-		 * JoinGroup was released along with SASL in 0.9.0. */
-		.feature = RD_KAFKA_FEATURE_SASL,
-		.depends = {
-			{ RD_KAFKAP_JoinGroup, 0, 0 },
-			{ -1 },
-		},
-	},
-	{
-		/* @brief >=0.10.0: SASL mechanism handshake (KIP-43) */
-		.feature = RD_KAFKA_FEATURE_SASL_HANDSHAKE,
-		.depends = {
-			{ RD_KAFKAP_SaslHandshake, 0, 0 },
-			{ -1 },
-		},
-	},
-	{ .feature = 0 }, /* sentinel */
+        },
+        {
+                /* @brief >=0.9.0: SASL (GSSAPI) authentication.
+                 * Since SASL is not using the Kafka protocol
+                 * we must use something else to map us to the
+                 * proper broker version support:
+                 * JoinGroup was released along with SASL in 0.9.0. */
+                .feature = RD_KAFKA_FEATURE_SASL_GSSAPI,
+                .depends = {
+                        { RD_KAFKAP_JoinGroup, 0, 0 },
+                        { -1 },
+                },
+        },
+        {
+                /* @brief >=0.10.0: SASL mechanism handshake (KIP-43)
+                 *                  to automatically support other mechanisms
+                 *                  than GSSAPI, such as PLAIN. */
+                .feature = RD_KAFKA_FEATURE_SASL_HANDSHAKE,
+                .depends = {
+                        { RD_KAFKAP_SaslHandshake, 0, 0 },
+                        { -1 },
+                },
+        },
+        {
+                /* @brief >=0.8.2: LZ4 compression.
+                 * Since LZ4 initially did not rely on a specific API
+                 * type or version (it does in >=0.10.0)
+                 * we must use something else to map us to the
+                 * proper broker version support:
+                 * GrooupCoordinator was released in 0.8.2 */
+                .feature = RD_KAFKA_FEATURE_LZ4,
+                .depends = {
+                        { RD_KAFKAP_GroupCoordinator, 0, 0 },
+                        { -1 },
+                },
+        },
+
+        { .feature = 0 }, /* sentinel */
 };
 
 
@@ -146,6 +163,12 @@ static const struct rd_kafka_feature_map {
  *        lists that corresponds to older broker versions.
  */
 
+/* >= 0.10.0.0: dummy for all future versions that support ApiVersionRequest */
+static struct rd_kafka_ApiVersion rd_kafka_ApiVersion_Queryable[] = {
+	{ RD_KAFKAP_ApiVersion, 0, 0 }
+};
+
+
 /* =~ 0.9.0 */
 static struct rd_kafka_ApiVersion rd_kafka_ApiVersion_0_9_0[] = {
 	{ RD_KAFKAP_Produce, 0, 1 },
@@ -218,6 +241,9 @@ int rd_kafka_get_legacy_ApiVersions (const char *broker_version,
 		_VERMAP("0.8.2", rd_kafka_ApiVersion_0_8_2),
 		_VERMAP("0.8.1", rd_kafka_ApiVersion_0_8_1),
 		_VERMAP("0.8.0", rd_kafka_ApiVersion_0_8_0),
+		{ "0.7.", NULL }, /* Unsupported */
+		{ "0.6.", NULL }, /* Unsupported */
+		_VERMAP("", rd_kafka_ApiVersion_Queryable),
 		{ NULL }
 	};
 	int i;
@@ -225,6 +251,8 @@ int rd_kafka_get_legacy_ApiVersions (const char *broker_version,
 
 	for (i = 0 ; vermap[i].pfx ; i++) {
 		if (!strncmp(vermap[i].pfx, broker_version, strlen(vermap[i].pfx))) {
+			if (!vermap[i].apis)
+				return 0;
 			*apisp = vermap[i].apis;
 			*api_cntp = vermap[i].api_cnt;
 			return 1;
@@ -243,6 +271,23 @@ int rd_kafka_get_legacy_ApiVersions (const char *broker_version,
 }
 
 
+/**
+ * @returns 1 if the provided broker version (probably)
+ *          supports api.version.request.
+ */
+int rd_kafka_ApiVersion_is_queryable (const char *broker_version) {
+	struct rd_kafka_ApiVersion *apis;
+	size_t api_cnt;
+
+
+	if (!rd_kafka_get_legacy_ApiVersions(broker_version,
+					     &apis, &api_cnt, 0))
+		return 0;
+
+	return apis == rd_kafka_ApiVersion_Queryable;
+}
+
+
 
 
 	
@@ -302,8 +347,7 @@ int rd_kafka_features_check (rd_kafka_broker_t *rkb,
 			r = rd_kafka_ApiVersion_check(broker_apis, broker_api_cnt,
 						      match);
 
-			rd_rkb_dbg(rkb, PROTOCOL | RD_KAFKA_DBG_BROKER,
-				   "APIVERSION",
+			rd_rkb_dbg(rkb, FEATURE, "APIVERSION",
 				   " Feature %s: %s (%hd..%hd) "
 				   "%ssupported by broker",
 				   rd_kafka_features2str(rd_kafka_feature_map[i].
@@ -315,8 +359,8 @@ int rd_kafka_features_check (rd_kafka_broker_t *rkb,
 			fails += !r;
 		}
 
-		rd_rkb_dbg(rkb, PROTOCOL | RD_KAFKA_DBG_BROKER,
-			   "APIVERSION", "%s feature %s",
+		rd_rkb_dbg(rkb, FEATURE, "APIVERSION",
+			   "%s feature %s",
 			   fails ? "Disabling" : "Enabling",
 			   rd_kafka_features2str(rd_kafka_feature_map[i].feature));
 
