@@ -38,31 +38,75 @@
  */
 #define RD_KAFKA_MSG_ATTR_GZIP             (1 << 0)
 #define RD_KAFKA_MSG_ATTR_SNAPPY           (1 << 1)
-#define RD_KAFKA_MSG_ATTR_LZ4              (1 << 2)
-#define RD_KAFKA_MSG_ATTR_COMPRESSION_MASK 0x7
+#define RD_KAFKA_MSG_ATTR_LZ4              (3)
+#define RD_KAFKA_MSG_ATTR_COMPRESSION_MASK 0x3
 #define RD_KAFKA_MSG_ATTR_CREATE_TIME      (0 << 3)
 #define RD_KAFKA_MSG_ATTR_LOG_APPEND_TIME  (1 << 3)
 
 
+
 typedef struct rd_kafka_msg_s {
+	rd_kafka_message_t rkm_rkmessage;  /* MUST be first field */
+#define rkm_len               rkm_rkmessage.len
+#define rkm_payload           rkm_rkmessage.payload
+#define rkm_opaque            rkm_rkmessage._private
+#define rkm_partition         rkm_rkmessage.partition
+#define rkm_offset            rkm_rkmessage.offset
+#define rkm_key               rkm_rkmessage.key
+#define rkm_key_len           rkm_rkmessage.key_len
+#define rkm_err               rkm_rkmessage.err
+
 	TAILQ_ENTRY(rd_kafka_msg_s)  rkm_link;
+
 	int        rkm_flags;
-	size_t     rkm_len;
-	void      *rkm_payload;
-	void      *rkm_opaque;
-	int32_t    rkm_partition;  /* partition specified */
-	rd_kafkap_bytes_t *rkm_key;
-        int64_t    rkm_offset;
+	/* @remark These additional flags must not collide with
+	 *         the RD_KAFKA_MSG_F_* flags in rdkafka.h */
+#define RD_KAFKA_MSG_F_FREE_RKM     0x10000 /* msg_t is allocated */
+#define RD_KAFKA_MSG_F_ACCOUNT      0x20000 /* accounted for in curr_msgs */
+
 	int64_t    rkm_timestamp;  /* Message format V1.
 				    * Meaning of timestamp depends on
 				    * message Attribute LogAppendtime (broker)
 				    * or CreateTime (producer).
 				    * Unit is milliseconds since epoch (UTC).*/
-	rd_ts_t    rkm_ts_timeout;
+	rd_kafka_timestamp_type_t rkm_tstype; /* rkm_timestamp type */
+
+	union {
+		struct {
+			rd_ts_t ts_timeout;
+		} producer;
+#define rkm_ts_timeout rkm_u.producer.ts_timeout
+	} rkm_u;
 } rd_kafka_msg_t;
 
 TAILQ_HEAD(rd_kafka_msg_head_s, rd_kafka_msg_s);
 
+
+
+/**
+ * @returns the message's total maximum on-wire size.
+ * @remark Depending on message version (MagicByte) the actual size
+ *         may be smaller.
+ */
+static RD_INLINE RD_UNUSED
+int32_t rd_kafka_msg_wire_size (const rd_kafka_msg_t *rkm) {
+	return (int32_t)(RD_KAFKAP_MESSAGE_OVERHEAD + rkm->rkm_len + rkm->rkm_key_len);
+}
+
+
+/**
+ * @returns the enveloping rd_kafka_msg_t pointer for a rd_kafka_msg_t
+ *          wrapped rd_kafka_message_t.
+ */
+static RD_INLINE RD_UNUSED
+rd_kafka_msg_t *rd_kafka_message2msg (rd_kafka_message_t *rkmessage) {
+	return (rd_kafka_msg_t *)rkmessage;
+}
+
+
+
+
+
 typedef struct rd_kafka_msgq_s {
 	TAILQ_HEAD(, rd_kafka_msg_s) rkmq_msgs;
 	rd_atomic32_t rkmq_msg_cnt;
@@ -82,6 +126,13 @@ static RD_INLINE RD_UNUSED int rd_kafka_msgq_len (rd_kafka_msgq_t *rkmq) {
 	return (int)rd_atomic32_get(&rkmq->rkmq_msg_cnt);
 }
 
+/**
+ * Returns the total number of bytes in the specified queue.
+ */
+static RD_INLINE RD_UNUSED size_t rd_kafka_msgq_size (rd_kafka_msgq_t *rkmq) {
+	return (size_t)rd_atomic64_get(&rkmq->rkmq_msg_bytes);
+}
+
 
 void rd_kafka_msg_destroy (rd_kafka_t *rk, rd_kafka_msg_t *rkm);
 
@@ -90,12 +141,18 @@ int rd_kafka_msg_new (rd_kafka_itopic_t *rkt, int32_t force_partition,
 		      char *payload, size_t len,
 		      const void *keydata, size_t keylen,
 		      void *msg_opaque);
+rd_kafka_msg_t *rd_kafka_msg_new00 (rd_kafka_itopic_t *rkt,
+				    int32_t partition,
+				    int msgflags,
+				    char *payload, size_t len,
+				    const void *key, size_t keylen,
+				    void *msg_opaque);
 
 
 static RD_INLINE RD_UNUSED void rd_kafka_msgq_init (rd_kafka_msgq_t *rkmq) {
 	TAILQ_INIT(&rkmq->rkmq_msgs);
-	rd_atomic32_set(&rkmq->rkmq_msg_cnt, 0);
-	rd_atomic64_set(&rkmq->rkmq_msg_bytes, 0);
+	rd_atomic32_init(&rkmq->rkmq_msg_cnt, 0);
+	rd_atomic64_init(&rkmq->rkmq_msg_bytes, 0);
 }
 
 /**
@@ -106,8 +163,8 @@ static RD_INLINE RD_UNUSED void rd_kafka_msgq_init (rd_kafka_msgq_t *rkmq) {
 static RD_INLINE RD_UNUSED void rd_kafka_msgq_concat (rd_kafka_msgq_t *dst,
 						   rd_kafka_msgq_t *src) {
 	TAILQ_CONCAT(&dst->rkmq_msgs, &src->rkmq_msgs, rkm_link);
-	(void)rd_atomic32_add(&dst->rkmq_msg_cnt, rd_atomic32_get(&src->rkmq_msg_cnt));
-	(void)rd_atomic64_add(&dst->rkmq_msg_bytes, rd_atomic64_get(&src->rkmq_msg_bytes));
+	rd_atomic32_add(&dst->rkmq_msg_cnt, rd_atomic32_get(&src->rkmq_msg_cnt));
+	rd_atomic64_add(&dst->rkmq_msg_bytes, rd_atomic64_get(&src->rkmq_msg_bytes));
 	rd_kafka_msgq_init(src);
 }
 
@@ -152,9 +209,10 @@ rd_kafka_msg_t *rd_kafka_msgq_deq (rd_kafka_msgq_t *rkmq,
 				   int do_count) {
 	if (likely(do_count)) {
 		rd_kafka_assert(NULL, rd_atomic32_get(&rkmq->rkmq_msg_cnt) > 0);
-		rd_kafka_assert(NULL, rd_atomic64_get(&rkmq->rkmq_msg_bytes) >= (int64_t)rkm->rkm_len);
-		(void)rd_atomic32_sub(&rkmq->rkmq_msg_cnt, 1);
-		(void)rd_atomic64_sub(&rkmq->rkmq_msg_bytes, rkm->rkm_len);
+		rd_kafka_assert(NULL, rd_atomic64_get(&rkmq->rkmq_msg_bytes) >= (int64_t)(rkm->rkm_len+rkm->rkm_key_len));
+		rd_atomic32_sub(&rkmq->rkmq_msg_cnt, 1);
+		rd_atomic64_sub(&rkmq->rkmq_msg_bytes,
+				rkm->rkm_len+rkm->rkm_key_len);
 	}
 
 	TAILQ_REMOVE(&rkmq->rkmq_msgs, rkm, rkm_link);
@@ -178,8 +236,8 @@ rd_kafka_msg_t *rd_kafka_msgq_pop (rd_kafka_msgq_t *rkmq) {
 static RD_INLINE RD_UNUSED void rd_kafka_msgq_insert (rd_kafka_msgq_t *rkmq,
 						   rd_kafka_msg_t *rkm) {
 	TAILQ_INSERT_HEAD(&rkmq->rkmq_msgs, rkm, rkm_link);
-	(void)rd_atomic32_add(&rkmq->rkmq_msg_cnt, 1);
-	(void)rd_atomic64_add(&rkmq->rkmq_msg_bytes, rkm->rkm_len);
+	rd_atomic32_add(&rkmq->rkmq_msg_cnt, 1);
+	rd_atomic64_add(&rkmq->rkmq_msg_bytes, rkm->rkm_len+rkm->rkm_key_len);
 }
 
 /**
@@ -188,8 +246,8 @@ static RD_INLINE RD_UNUSED void rd_kafka_msgq_insert (rd_kafka_msgq_t *rkmq,
 static RD_INLINE RD_UNUSED void rd_kafka_msgq_enq (rd_kafka_msgq_t *rkmq,
 						rd_kafka_msg_t *rkm) {
 	TAILQ_INSERT_TAIL(&rkmq->rkmq_msgs, rkm, rkm_link);
-	(void)rd_atomic32_add(&rkmq->rkmq_msg_cnt, 1);
-	(void)rd_atomic64_add(&rkmq->rkmq_msg_bytes, rkm->rkm_len);
+	rd_atomic32_add(&rkmq->rkmq_msg_cnt, 1);
+	rd_atomic64_add(&rkmq->rkmq_msg_bytes, rkm->rkm_len+rkm->rkm_key_len);
 }
 
 
