@@ -15,6 +15,10 @@ void rd_kafka_yield (rd_kafka_t *rk) {
 void rd_kafka_q_destroy_final (rd_kafka_q_t *rkq) {
 
         mtx_lock(&rkq->rkq_lock);
+	if (unlikely(rkq->rkq_qio != NULL)) {
+		rd_free(rkq->rkq_qio);
+		rkq->rkq_qio = NULL;
+	}
         rd_kafka_q_fwd_set0(rkq, NULL, 0/*no-lock*/);
         rd_kafka_q_disable0(rkq, 0/*no-lock*/);
         rd_kafka_q_purge0(rkq, 0/*no-lock*/);
@@ -38,6 +42,7 @@ void rd_kafka_q_init (rd_kafka_q_t *rkq, rd_kafka_t *rk) {
         rkq->rkq_refcnt = 1;
         rkq->rkq_flags  = RD_KAFKA_Q_F_READY;
         rkq->rkq_rk     = rk;
+	rkq->rkq_qio    = NULL;
 	mtx_init(&rkq->rkq_lock, mtx_plain);
 	cnd_init(&rkq->rkq_cond);
 }
@@ -46,10 +51,15 @@ void rd_kafka_q_init (rd_kafka_q_t *rkq, rd_kafka_t *rk) {
 /**
  * Allocate a new queue and initialize it.
  */
-rd_kafka_q_t *rd_kafka_q_new (rd_kafka_t *rk) {
+rd_kafka_q_t *rd_kafka_q_new0 (rd_kafka_t *rk, const char *func, int line) {
         rd_kafka_q_t *rkq = rd_malloc(sizeof(*rkq));
         rd_kafka_q_init(rkq, rk);
         rkq->rkq_flags |= RD_KAFKA_Q_F_ALLOCATED;
+#if ENABLE_DEVEL
+	rd_snprintf(rkq->rkq_name, sizeof(rkq->rkq_name), "%s:%d", func, line);
+#else
+	rkq->rkq_name = func;
+#endif
         return rkq;
 }
 
@@ -72,7 +82,6 @@ void rd_kafka_q_fwd_set0 (rd_kafka_q_t *srcq, rd_kafka_q_t *destq,
 	}
 	if (destq) {
 		rd_kafka_q_keep(destq);
-		srcq->rkq_fwdq = destq;
 
 		/* If rkq has ops in queue, append them to fwdq's queue.
 		 * This is an irreversible operation. */
@@ -80,6 +89,8 @@ void rd_kafka_q_fwd_set0 (rd_kafka_q_t *srcq, rd_kafka_q_t *destq,
 			rd_dassert(destq->rkq_flags & RD_KAFKA_Q_F_READY);
 			rd_kafka_q_concat(destq, srcq);
 		}
+
+		srcq->rkq_fwdq = destq;
 	}
         if (do_lock)
                 mtx_unlock(&srcq->rkq_lock);
@@ -185,6 +196,9 @@ int rd_kafka_q_move_cnt (rd_kafka_q_t *dstq, rd_kafka_q_t *srcq,
 	}
 
 	if (!dstq->rkq_fwdq && !srcq->rkq_fwdq) {
+		if (cnt > 0 && dstq->rkq_qlen == 0)
+			rd_kafka_q_io_event(dstq);
+
 		/* Optimization, if 'cnt' is equal/larger than all
 		 * items of 'srcq' we can move the entire queue. */
 		if (cnt == -1 ||
@@ -226,14 +240,12 @@ int rd_kafka_q_move_cnt (rd_kafka_q_t *dstq, rd_kafka_q_t *srcq,
  * Filters out outdated ops.
  */
 static RD_INLINE rd_kafka_op_t *rd_kafka_op_filter (rd_kafka_q_t *rkq,
-                                                   rd_kafka_op_t *rko) {
+						    rd_kafka_op_t *rko,
+						    int version) {
         if (unlikely(!rko))
                 return NULL;
 
-        if (unlikely(rko->rko_version && rko->rko_rktp &&
-                     rko->rko_version <
-                     rd_atomic32_get(&rd_kafka_toppar_s2i(rko->rko_rktp)->
-                                     rktp_version))) {
+        if (unlikely(rd_kafka_op_version_outdated(rko, version))) {
 		rd_kafka_q_deq0(rkq, rko);
                 rd_kafka_op_destroy(rko);
                 return NULL;
@@ -249,8 +261,23 @@ static RD_INLINE rd_kafka_op_t *rd_kafka_op_filter (rd_kafka_q_t *rkq,
  *
  * Locality: any thread.
  */
-rd_kafka_op_t *rd_kafka_q_pop (rd_kafka_q_t *rkq, int timeout_ms,
-                               int32_t version) {
+
+
+/**
+ * Serve q like rd_kafka_q_serve() until an op is found that can be returned
+ * as an event to the application.
+ *
+ * @returns the first event:able op, or NULL on timeout.
+ *
+ * Locality: any thread
+ */
+rd_kafka_op_t *rd_kafka_q_pop_serve (rd_kafka_q_t *rkq, int timeout_ms,
+				     int32_t version, int cb_type,
+				     int (*callback) (rd_kafka_t *rk,
+						      rd_kafka_op_t *rko,
+						      int cb_type,
+						      void *opaque),
+				     void *opaque) {
 	rd_kafka_op_t *rko;
 
 	if (timeout_ms == RD_POLL_INFINITE)
@@ -262,13 +289,30 @@ rd_kafka_op_t *rd_kafka_q_pop (rd_kafka_q_t *rkq, int timeout_ms,
                 do {
                         /* Filter out outdated ops */
                         while ((rko = TAILQ_FIRST(&rkq->rkq_q)) &&
-                               !(rko = rd_kafka_op_filter(rkq, rko)))
+                               !(rko = rd_kafka_op_filter(rkq, rko, version)))
                                 ;
 
                         if (rko) {
+				int handled;
+
                                 /* Proper versioned op */
                                 rd_kafka_q_deq0(rkq, rko);
-                                break;
+
+				/* Ops with callbacks are considered handled
+				 * and we move on to the next op, if any.
+				 * Ops w/o callbacks are returned immediately */
+				if (callback) {
+					handled = callback(rkq->rkq_rk, rko,
+							   cb_type, opaque);
+					if (handled) {
+						rd_kafka_op_destroy(rko);
+						rko = NULL;
+					}
+				} else
+					handled = 0;
+
+				if (!handled)
+					break;
                         }
 
                         /* No op, wait for one */
@@ -295,7 +339,8 @@ rd_kafka_op_t *rd_kafka_q_pop (rd_kafka_q_t *rkq, int timeout_ms,
                 /* Since the q_pop may block we need to release the parent
                  * queue's lock. */
                 mtx_unlock(&rkq->rkq_lock);
-		rko = rd_kafka_q_pop(fwdq, timeout_ms, version);
+		rko = rd_kafka_q_pop_serve(fwdq, timeout_ms, version,
+					   cb_type, callback, opaque);
                 rd_kafka_q_destroy(fwdq);
         }
 
@@ -303,6 +348,11 @@ rd_kafka_op_t *rd_kafka_q_pop (rd_kafka_q_t *rkq, int timeout_ms,
 	return rko;
 }
 
+rd_kafka_op_t *rd_kafka_q_pop (rd_kafka_q_t *rkq, int timeout_ms,
+                               int32_t version) {
+	return rd_kafka_q_pop_serve(rkq, timeout_ms, version, 0, NULL, NULL);
+}
+
 
 /**
  * Pop all available ops from a queue and call the provided 
@@ -316,8 +366,7 @@ rd_kafka_op_t *rd_kafka_q_pop (rd_kafka_q_t *rkq, int timeout_ms,
 int rd_kafka_q_serve (rd_kafka_q_t *rkq, int timeout_ms,
                       int max_cnt, int cb_type,
                       int (*callback) (rd_kafka_t *rk, rd_kafka_op_t *rko,
-                                       int cb_type,
-                                              void *opaque),
+                                       int cb_type, void *opaque),
                       void *opaque) {
         rd_kafka_t *rk = rkq->rkq_rk;
 	rd_kafka_op_t *rko;
@@ -401,49 +450,100 @@ void rd_kafka_message_destroy (rd_kafka_message_t *rkmessage) {
 
 	if (likely((rko = (rd_kafka_op_t *)rkmessage->_private) != NULL))
 		rd_kafka_op_destroy(rko);
-	else
-		rd_free(rkmessage);
+	else {
+		rd_kafka_msg_t *rkm = rd_kafka_message2msg(rkmessage);
+		rd_kafka_msg_destroy(NULL, rkm);
+	}
 }
 
 
 rd_kafka_message_t *rd_kafka_message_new (void) {
-        rd_kafka_message_t *rkmessage;
-        rkmessage = rd_calloc(1, sizeof(*rkmessage));
-        return rkmessage;
+        rd_kafka_msg_t *rkm = rd_calloc(1, sizeof(*rkm));
+        return (rd_kafka_message_t *)rkm;
 }
 
-rd_kafka_message_t *rd_kafka_message_get (rd_kafka_op_t *rko) {
-	rd_kafka_message_t *rkmessage;
 
-	if (rko) {
-		rkmessage = &rko->rko_rkmessage;
+static rd_kafka_message_t *
+rd_kafka_message_setup (rd_kafka_op_t *rko, rd_kafka_message_t *rkmessage) {
+	rd_kafka_itopic_t *rkt;
+	rd_kafka_toppar_t *rktp = NULL;
+
+	if (rko->rko_type == RD_KAFKA_OP_DR) {
+		rkt = rd_kafka_topic_s2i(rko->rko_u.dr.s_rkt);
+	} else {
+		if (rko->rko_rktp) {
+			rktp = rd_kafka_toppar_s2i(rko->rko_rktp);
+			rkt = rktp->rktp_rkt;
+		} else
+			rkt = NULL;
+
 		rkmessage->_private = rko;
+	}
 
-		if (!rkmessage->rkt && rko->rko_rktp)
-			rkmessage->rkt =
-				rd_kafka_topic_keep_a(
-					rd_kafka_toppar_s2i(rko->rko_rktp)->
-					rktp_rkt);
-	} else
-                rkmessage = rd_kafka_message_new();
+
+	if (!rkmessage->rkt && rkt)
+		rkmessage->rkt = rd_kafka_topic_keep_a(rkt);
+
+	if (rktp)
+		rkmessage->partition = rktp->rktp_partition;
+
+	if (!rkmessage->err)
+		rkmessage->err = rko->rko_err;
 
 	return rkmessage;
 }
 
 
-int64_t rd_kafka_message_timestamp (const rd_kafka_message_t *rkmessage,
-				    rd_kafka_timestamp_type_t *tstype) {
-	const rd_kafka_op_t *rko = rkmessage->_private;
 
-	if (!rko || rko->rko_err) {
-		*tstype = RD_KAFKA_TIMESTAMP_NOT_AVAILABLE;
-		return -1;;
+rd_kafka_message_t *rd_kafka_message_get_from_rkm (rd_kafka_op_t *rko,
+						   rd_kafka_msg_t *rkm) {
+	return rd_kafka_message_setup(rko, &rkm->rkm_rkmessage);
+}
+
+rd_kafka_message_t *rd_kafka_message_get (rd_kafka_op_t *rko) {
+	rd_kafka_message_t *rkmessage;
+
+	if (!rko)
+		return rd_kafka_message_new(); /* empty */
+
+	switch (rko->rko_type)
+	{
+	case RD_KAFKA_OP_FETCH:
+		/* Use embedded rkmessage */
+		rkmessage = &rko->rko_u.fetch.rkm.rkm_rkmessage;
+		break;
+
+	case RD_KAFKA_OP_ERR:
+	case RD_KAFKA_OP_CONSUMER_ERR:
+		rkmessage = &rko->rko_u.err.rkm.rkm_rkmessage;
+		rkmessage->payload = rko->rko_u.err.errstr;
+		rkmessage->offset  = rko->rko_u.err.offset;
+		break;
+
+	default:
+		rd_kafka_assert(NULL, !*"unhandled optype");
+		RD_NOTREACHED();
+		return NULL;
 	}
 
+	return rd_kafka_message_setup(rko, rkmessage);
+}
 
-	*tstype = rko->rko_tstype;
 
-	return rko->rko_timestamp;
+int64_t rd_kafka_message_timestamp (const rd_kafka_message_t *rkmessage,
+				    rd_kafka_timestamp_type_t *tstype) {
+	rd_kafka_msg_t *rkm;
+
+	if (rkmessage->err) {
+		*tstype = RD_KAFKA_TIMESTAMP_NOT_AVAILABLE;
+		return -1;
+	}
+
+	rkm = rd_kafka_message2msg((rd_kafka_message_t *)rkmessage);
+
+	*tstype = rkm->rkm_tstype;
+
+	return rkm->rkm_timestamp;
 }
 
 
@@ -497,10 +597,7 @@ int rd_kafka_q_serve_rkmessages (rd_kafka_q_t *rkq, int timeout_ms,
 
                 mtx_unlock(&rkq->rkq_lock);
 
-                if (rko->rko_version && rko->rko_rktp &&
-                    rko->rko_version <
-                    rd_atomic32_get(&rd_kafka_toppar_s2i(rko->rko_rktp)->
-                                    rktp_version)) {
+		if (rd_kafka_op_version_outdated(rko, 0)) {
                         /* Outdated op, put on discard queue */
                         TAILQ_INSERT_TAIL(&tmpq, rko, rko_link);
                         continue;
@@ -514,15 +611,15 @@ int rd_kafka_q_serve_rkmessages (rd_kafka_q_t *rkq, int timeout_ms,
                 }
 
 		/* Auto-commit offset, if enabled. */
-		if (!rko->rko_err) {
+		if (!rko->rko_err && rko->rko_type == RD_KAFKA_OP_FETCH) {
                         rd_kafka_toppar_t *rktp;
                         rktp = rd_kafka_toppar_s2i(rko->rko_rktp);
 			rd_kafka_toppar_lock(rktp);
-			rktp->rktp_app_offset = rko->rko_offset+1;
+			rktp->rktp_app_offset = rko->rko_u.fetch.rkm.rkm_offset+1;
                         if (rktp->rktp_cgrp &&
 			    rk->rk_conf.enable_auto_offset_store)
                                 rd_kafka_offset_store0(rktp,
-                                                       rko->rko_offset+1,
+						       rktp->rktp_app_offset,
                                                        0/* no lock */);
 			rd_kafka_toppar_unlock(rktp);
                 }
@@ -546,24 +643,18 @@ int rd_kafka_q_serve_rkmessages (rd_kafka_q_t *rkq, int timeout_ms,
 
 
 void rd_kafka_queue_destroy (rd_kafka_queue_t *rkqu) {
-        int do_destroy;
-
-        mtx_lock(&rkqu->rkqu_q.rkq_lock);
-        do_destroy = rkqu->rkqu_q.rkq_refcnt == 1;
-        mtx_unlock(&rkqu->rkqu_q.rkq_lock);
-	rd_kafka_q_destroy(&rkqu->rkqu_q);
-        if (!do_destroy)
-		return; /* Still references */
-
+	rd_kafka_q_disable(rkqu->rkqu_q);
+	rd_kafka_q_destroy(rkqu->rkqu_q);
 	rd_free(rkqu);
 }
 
-rd_kafka_queue_t *rd_kafka_queue_new (rd_kafka_t *rk) {
+rd_kafka_queue_t *rd_kafka_queue_new0 (rd_kafka_t *rk, rd_kafka_q_t *rkq) {
 	rd_kafka_queue_t *rkqu;
 
 	rkqu = rd_calloc(1, sizeof(*rkqu));
 
-	rd_kafka_q_init(&rkqu->rkqu_q, rk);
+	rkqu->rkqu_q = rkq;
+	rd_kafka_q_keep(rkq);
 
         rkqu->rkqu_rk = rk;
 
@@ -571,6 +662,64 @@ rd_kafka_queue_t *rd_kafka_queue_new (rd_kafka_t *rk) {
 }
 
 
+rd_kafka_queue_t *rd_kafka_queue_new (rd_kafka_t *rk) {
+	rd_kafka_q_t *rkq;
+	rd_kafka_queue_t *rkqu;
+
+	rkq = rd_kafka_q_new(rk);
+	rkqu = rd_kafka_queue_new0(rk, rkq);
+	rd_kafka_q_destroy(rkq); /* Loose refcount from q_new, one is held
+				  * by queue_new0 */
+	return rkqu;
+}
+
+
+rd_kafka_queue_t *rd_kafka_queue_get_main (rd_kafka_t *rk) {
+	return rd_kafka_queue_new0(rk, rk->rk_rep);
+}
+
+
+rd_kafka_queue_t *rd_kafka_queue_get_consumer (rd_kafka_t *rk) {
+	if (!rk->rk_cgrp)
+		return NULL;
+	return rd_kafka_queue_new0(rk, rk->rk_cgrp->rkcg_q);
+}
+
+void rd_kafka_queue_forward (rd_kafka_queue_t *src, rd_kafka_queue_t *dst) {
+	rd_kafka_q_fwd_set(src->rkqu_q, dst ? dst->rkqu_q : NULL);
+}
+
+
+size_t rd_kafka_queue_length (rd_kafka_queue_t *rkqu) {
+	return (size_t)rd_kafka_q_len(rkqu->rkqu_q);
+}
+
+void rd_kafka_queue_io_event_enable (rd_kafka_queue_t *rkqu, int fd,
+				     const void *payload, size_t size) {
+	rd_kafka_q_t *rkq = rkqu->rkqu_q;
+	struct rd_kafka_q_io *qio;
+
+	if (fd != -1) {
+		qio = rd_malloc(sizeof(*qio) + size);
+		qio->fd = fd;
+		qio->size = size;
+		qio->payload = (void *)(qio+1);
+		memcpy(qio->payload, payload, size);
+	}
+
+	mtx_lock(&rkq->rkq_lock);
+	if (rkq->rkq_qio) {
+		rd_free(rkq->rkq_qio);
+		rkq->rkq_qio = NULL;
+	}
+
+	if (fd != -1) {
+		rkq->rkq_qio = qio;
+	}
+
+	mtx_unlock(&rkq->rkq_lock);
+}
+
 
 /**
  * Helper: wait for single op on 'rkq', and return its error,
@@ -592,6 +741,40 @@ rd_kafka_resp_err_t rd_kafka_q_wait_result (rd_kafka_q_t *rkq, int timeout_ms) {
 }
 
 
+/**
+ * Apply \p callback on each op in queue.
+ * If the callback wishes to remove the rko it must do so using
+ * using rd_kafka_op_deq0().
+ *
+ * @returns the sum of \p callback() return values.
+ * @remark rkq will be locked, callers should take care not to
+ *         interact with \p rkq through other means from the callback to avoid
+ *         deadlocks.
+ */
+int rd_kafka_q_apply (rd_kafka_q_t *rkq,
+                      int (*callback) (rd_kafka_q_t *rkq, rd_kafka_op_t *rko,
+                                       void *opaque),
+                      void *opaque) {
+	rd_kafka_op_t *rko, *next;
+        int cnt = 0;
+
+        mtx_lock(&rkq->rkq_lock);
+        if (rkq->rkq_fwdq) {
+		cnt = rd_kafka_q_apply(rkq->rkq_fwdq, callback, opaque);
+                mtx_unlock(&rkq->rkq_lock);
+		return cnt;
+	}
+
+	next = TAILQ_FIRST(&rkq->rkq_q);
+	while ((rko = next)) {
+		next = TAILQ_NEXT(next, rko_link);
+                cnt += callback(rkq, rko, opaque);
+	}
+        mtx_unlock(&rkq->rkq_lock);
+
+        return cnt;
+}
+
 /**
  * @brief Convert relative to absolute offsets and also purge any messages
  *        that are older than \p min_offset.
@@ -613,7 +796,12 @@ void rd_kafka_q_fix_offsets (rd_kafka_q_t *rkq, int64_t min_offset,
 	while ((rko = next)) {
 		next = TAILQ_NEXT(next, rko_link);
 
-		if (rko->rko_offset < min_offset &&
+		if (unlikely(rko->rko_type != RD_KAFKA_OP_FETCH))
+			continue;
+
+		rko->rko_u.fetch.rkm.rkm_offset += base_offset;
+
+		if (rko->rko_u.fetch.rkm.rkm_offset < min_offset &&
 		    rko->rko_err != RD_KAFKA_RESP_ERR__NOT_IMPLEMENTED) {
 			adj_len++;
 			adj_size += rko->rko_len;
@@ -621,8 +809,6 @@ void rd_kafka_q_fix_offsets (rd_kafka_q_t *rkq, int64_t min_offset,
 			rd_kafka_op_destroy(rko);
 			continue;
 		}
-
-		rko->rko_offset += base_offset;
 	}
 
 
