@@ -457,16 +457,13 @@ static RD_UNUSED int rd_kafka_sasl_cb_canon (sasl_conn_t *conn,
 					     unsigned *out_len) {
 	rd_kafka_transport_t *rktrans = context;
 
-	if (strstr(rktrans->rktrans_rkb->rkb_rk->rk_conf.
-		   sasl.mechanisms, "GSSAPI")) {
-		*out_len = rd_snprintf(out, out_max, "%s",
-				       rktrans->rktrans_rkb->rkb_rk->
-				       rk_conf.sasl.principal);
-	} else if (!strcmp(rktrans->rktrans_rkb->rkb_rk->rk_conf.
-			   sasl.mechanisms, "PLAIN")) {
-		*out_len = rd_snprintf(out, out_max, "%.*s", inlen, in);
-	} else
-		out = NULL;
+        if (strstr(rktrans->rktrans_rkb->rkb_rk->rk_conf.
+                   sasl.mechanisms, "GSSAPI"))
+                *out_len = rd_snprintf(out, out_max, "%s",
+                                       rktrans->rktrans_rkb->rkb_rk->
+                                       rk_conf.sasl.principal);
+        else
+                *out_len = rd_snprintf(out, out_max, "%.*s", inlen, in);
 
 	rd_rkb_dbg(rktrans->rktrans_rkb, SECURITY, "LIBSASL",
 		   "CB_CANON: flags 0x%x, \"%.*s\" @ \"%s\": returning \"%.*s\"",
@@ -500,8 +497,29 @@ int rd_kafka_sasl_client_new (rd_kafka_transport_t *rktrans,
 		{ SASL_CB_LIST_END }
 	};
 
+        /* Verify broker support:
+         * - RD_KAFKA_FEATURE_SASL_GSSAPI - GSSAPI supported
+         * - RD_KAFKA_FEATURE_SASL_HANDSHAKE - GSSAPI, PLAIN and possibly
+         *   other mechanisms supported. */
+        if (!strcmp(rk->rk_conf.sasl.mechanisms, "GSSAPI")) {
+                if (!(rkb->rkb_features & RD_KAFKA_FEATURE_SASL_GSSAPI)) {
+                        rd_snprintf(errstr, errstr_size,
+                                    "SASL GSSAPI authentication not supported "
+                                    "by broker");
+                        return -1;
+                }
+        } else if (!(rkb->rkb_features & RD_KAFKA_FEATURE_SASL_HANDSHAKE)) {
+                rd_snprintf(errstr, errstr_size,
+                            "SASL Handshake not supported by broker "
+                            "(required by mechanism %s)%s",
+                            rk->rk_conf.sasl.mechanisms,
+                            rk->rk_conf.api_version_request ? "" :
+                            ": try api.version.request=true");
+                return -1;
+        }
+
 	/* SASL_CB_USER is needed for PLAIN but breaks GSSAPI */
-	if (!strcmp(rk->rk_conf.sasl.service_name, "PLAIN")) {
+	if (strcmp(rk->rk_conf.sasl.mechanisms, "GSSAPI")) {
 		int endidx;
 		/* Find end of callbacks array */
 		for (endidx = 0 ;
@@ -510,6 +528,7 @@ int rd_kafka_sasl_client_new (rd_kafka_transport_t *rktrans,
 
 		callbacks[endidx].id = SASL_CB_USER;
 		callbacks[endidx].proc = (void *)rd_kafka_sasl_cb_getsimple;
+		callbacks[endidx].context = rktrans;
 		endidx++;
 		callbacks[endidx].id = SASL_CB_LIST_END;
 	}
@@ -612,7 +631,7 @@ void rd_kafka_broker_sasl_init (rd_kafka_broker_t *rkb) {
 		return; /* kinit not configured, no need to start timer */
 
 	rd_kafka_timer_start(&rk->rk_timers, &rkb->rkb_sasl_kinit_refresh_tmr,
-			     rk->rk_conf.sasl.relogin_min_time * 1000,
+			     rk->rk_conf.sasl.relogin_min_time * 1000ll,
 			     rd_kafka_sasl_kinit_refresh_tmr_cb, rkb);
 }
 
@@ -620,42 +639,46 @@ void rd_kafka_broker_sasl_init (rd_kafka_broker_t *rkb) {
 
 int rd_kafka_sasl_conf_validate (rd_kafka_t *rk,
 				 char *errstr, size_t errstr_size) {
-	rd_kafka_broker_t rkb;
-	char *cmd;
-	char tmperr[128];
 
 	if (strcmp(rk->rk_conf.sasl.mechanisms, "GSSAPI"))
 		return 0;
 
-	memset(&rkb, 0, sizeof(rkb));
-	strcpy(rkb.rkb_nodename, "ATestBroker:9092");
-	rkb.rkb_rk = rk;
-	mtx_init(&rkb.rkb_lock, mtx_plain);
+	if (rk->rk_conf.sasl.kinit_cmd) {
+		rd_kafka_broker_t rkb;
+		char *cmd;
+		char tmperr[128];
 
-	cmd = rd_string_render(rk->rk_conf.sasl.kinit_cmd,
-			       tmperr, sizeof(tmperr),
-			       render_callback, &rkb);
+		memset(&rkb, 0, sizeof(rkb));
+		strcpy(rkb.rkb_nodename, "ATestBroker:9092");
+		rkb.rkb_rk = rk;
+		mtx_init(&rkb.rkb_lock, mtx_plain);
 
-	mtx_destroy(&rkb.rkb_lock);
+		cmd = rd_string_render(rk->rk_conf.sasl.kinit_cmd,
+				       tmperr, sizeof(tmperr),
+				       render_callback, &rkb);
 
-	if (!cmd) {
-		rd_snprintf(errstr, errstr_size,
-			    "Invalid sasl.kerberos.kinit.cmd value: %s",
-			    tmperr);
-		return -1;
+		mtx_destroy(&rkb.rkb_lock);
+
+		if (!cmd) {
+			rd_snprintf(errstr, errstr_size,
+				    "Invalid sasl.kerberos.kinit.cmd value: %s",
+				    tmperr);
+			return -1;
+		}
+
+		rd_free(cmd);
 	}
 
-	rd_free(cmd);
 	return 0;
 }
 
 
 /**
  * Global SASL termination.
- * NOTE: Should not be called since the application may be using SASL too.
  */
 void rd_kafka_sasl_global_term (void) {
-	sasl_done();
+	/* NOTE: Should not be called since the application may be using SASL too*/
+	/* sasl_done(); */
 	mtx_destroy(&rd_kafka_sasl_kinit_lock);
 }
 
