@@ -48,7 +48,7 @@ typedef struct rd_kafka_timer_s {
 	TAILQ_ENTRY(rd_kafka_timer_s)  rtmr_link;
 
 	rd_ts_t rtmr_next;
-	int     rtmr_interval;   /* interval in microseconds */
+	rd_ts_t rtmr_interval;   /* interval in microseconds */
 
 	void  (*rtmr_callback) (rd_kafka_timers_t *rkts, void *arg);
 	void   *rtmr_arg;
@@ -59,7 +59,7 @@ typedef struct rd_kafka_timer_s {
 void rd_kafka_timer_stop (rd_kafka_timers_t *rkts,
                           rd_kafka_timer_t *rtmr, int lock);
 void rd_kafka_timer_start (rd_kafka_timers_t *rkts,
-			   rd_kafka_timer_t *rtmr, int interval,
+			   rd_kafka_timer_t *rtmr, rd_ts_t interval,
 			   void (*callback) (rd_kafka_timers_t *rkts,
                                              void *arg),
 			   void *arg);
