@@ -126,7 +126,58 @@ static inline bool is_little_endian(void)
 	return false;
 }
 
-#ifndef _MSC_VER
+#if defined(__xlc__) // xlc compiler on AIX
+#define rd_clz(n)   __cntlz4(n)
+#define rd_ctz(n)   __cnttz4(n)
+#define rd_ctz64(n) __cnttz8(n)
+
+#elif defined(__SUNPRO_C) // Solaris Studio compiler on sun  
+/*
+ * Source for following definitions is Hacker’s Delight, Second Edition by Henry S. Warren
+ * http://www.hackersdelight.org/permissions.htm
+ */
+u32 rd_clz(u32 x) {
+   u32 n;
+
+   if (x == 0) return(32);
+   n = 1;
+   if ((x >> 16) == 0) {n = n +16; x = x <<16;}
+   if ((x >> 24) == 0) {n = n + 8; x = x << 8;}
+   if ((x >> 28) == 0) {n = n + 4; x = x << 4;}
+   if ((x >> 30) == 0) {n = n + 2; x = x << 2;}
+   n = n - (x >> 31);
+   return n;
+}
+
+u32 rd_ctz(u32 x) {
+   u32 y;
+   u32 n;
+
+   if (x == 0) return 32;
+   n = 31;
+   y = x <<16;  if (y != 0) {n = n -16; x = y;}
+   y = x << 8;  if (y != 0) {n = n - 8; x = y;}
+   y = x << 4;  if (y != 0) {n = n - 4; x = y;}
+   y = x << 2;  if (y != 0) {n = n - 2; x = y;}
+   y = x << 1;  if (y != 0) {n = n - 1;}
+   return n;
+}
+
+u64 rd_ctz64(u64 x) {
+   u64 y;
+   u64 n;
+
+   if (x == 0) return 64;
+   n = 63;
+   y = x <<32;  if (y != 0) {n = n -32; x = y;}
+   y = x <<16;  if (y != 0) {n = n -16; x = y;}
+   y = x << 8;  if (y != 0) {n = n - 8; x = y;}
+   y = x << 4;  if (y != 0) {n = n - 4; x = y;}
+   y = x << 2;  if (y != 0) {n = n - 2; x = y;}
+   y = x << 1;  if (y != 0) {n = n - 1;}
+   return n;
+}
+#elif !defined(_MSC_VER)
 #define rd_clz(n)   __builtin_clz(n)
 #define rd_ctz(n)   __builtin_ctz(n)
 #define rd_ctz64(n) __builtin_ctzll(n)
@@ -170,12 +221,12 @@ static inline int log2_floor(u32 n)
 	return n == 0 ? -1 : 31 ^ rd_clz(n);
 }
 
-static inline int find_lsb_set_non_zero(u32 n)
+static inline RD_UNUSED int find_lsb_set_non_zero(u32 n)
 {
 	return rd_ctz(n);
 }
 
-static inline int find_lsb_set_non_zero64(u64 n)
+static inline RD_UNUSED int find_lsb_set_non_zero64(u64 n)
 {
 	return rd_ctz64(n);
 }
@@ -580,11 +631,11 @@ static inline u32 hash(const char *p, int shift)
  *
  * This last factor dominates the blowup, so the final estimate is:
  */
-size_t snappy_max_compressed_length(size_t source_len)
+size_t rd_kafka_snappy_max_compressed_length(size_t source_len)
 {
 	return 32 + source_len + source_len / 6;
 }
-EXPORT_SYMBOL(snappy_max_compressed_length);
+EXPORT_SYMBOL(rd_kafka_snappy_max_compressed_length);
 
 enum {
 	LITERAL = 0,
@@ -685,14 +736,14 @@ static inline char *emit_copy(char *op, int offset, int len)
 }
 
 /**
- * snappy_uncompressed_length - return length of uncompressed output.
+ * rd_kafka_snappy_uncompressed_length - return length of uncompressed output.
  * @start: compressed buffer
  * @n: length of compressed buffer.
  * @result: Write the length of the uncompressed output here.
  *
  * Returns true when successfull, otherwise false.
  */
-bool snappy_uncompressed_length(const char *start, size_t n, size_t * result)
+bool rd_kafka_snappy_uncompressed_length(const char *start, size_t n, size_t * result)
 {
 	u32 v = 0;
 	const char *limit = start + n;
@@ -703,7 +754,7 @@ bool snappy_uncompressed_length(const char *start, size_t n, size_t * result)
 		return false;
 	}
 }
-EXPORT_SYMBOL(snappy_uncompressed_length);
+EXPORT_SYMBOL(rd_kafka_snappy_uncompressed_length);
 
 /*
  * The size of a compression block. Note that many parts of the compression
@@ -1404,7 +1455,7 @@ static inline int sn_compress(struct snappy_env *env, struct source *reader,
 
 		/* Compress input_fragment and append to dest */
 		char *dest;
-		dest = sink_peek(writer, snappy_max_compressed_length(num_to_read));
+		dest = sink_peek(writer, rd_kafka_snappy_max_compressed_length(num_to_read));
 		if (!dest) {
 			/*
 			 * Need a scratch buffer for the output,
@@ -1429,7 +1480,7 @@ out:
 
 #ifdef SG
 
-int snappy_compress_iov(struct snappy_env *env,
+int rd_kafka_snappy_compress_iov(struct snappy_env *env,
 			struct iovec *iov_in,
 			int iov_in_len,
 			size_t input_length,
@@ -1454,10 +1505,10 @@ int snappy_compress_iov(struct snappy_env *env,
 	*compressed_length = writer.written;
 	return err;
 }
-EXPORT_SYMBOL(snappy_compress_iov);
+EXPORT_SYMBOL(rd_kafka_snappy_compress_iov);
 
 /**
- * snappy_compress - Compress a buffer using the snappy compressor.
+ * rd_kafka_snappy_compress - Compress a buffer using the snappy compressor.
  * @env: Preallocated environment
  * @input: Input buffer
  * @input_length: Length of input_buffer
@@ -1467,13 +1518,13 @@ EXPORT_SYMBOL(snappy_compress_iov);
  * Return 0 on success, otherwise an negative error code.
  *
  * The output buffer must be at least
- * snappy_max_compressed_length(input_length) bytes long.
+ * rd_kafka_snappy_max_compressed_length(input_length) bytes long.
  *
- * Requires a preallocated environment from snappy_init_env.
+ * Requires a preallocated environment from rd_kafka_snappy_init_env.
  * The environment does not keep state over individual calls
  * of this function, just preallocates the memory.
  */
-int snappy_compress(struct snappy_env *env,
+int rd_kafka_snappy_compress(struct snappy_env *env,
 		    const char *input,
 		    size_t input_length,
 		    char *compressed, size_t *compressed_length)
@@ -1487,13 +1538,13 @@ int snappy_compress(struct snappy_env *env,
 		.iov_len = 0xffffffff,
 	};
 	int out = 1;
-	return snappy_compress_iov(env, 
+	return rd_kafka_snappy_compress_iov(env, 
 				   &iov_in, 1, input_length, 
 				   &iov_out, &out, compressed_length);
 }
-EXPORT_SYMBOL(snappy_compress);
+EXPORT_SYMBOL(rd_kafka_snappy_compress);
 
-int snappy_uncompress_iov(struct iovec *iov_in, int iov_in_len,
+int rd_kafka_snappy_uncompress_iov(struct iovec *iov_in, int iov_in_len,
 			   size_t input_len, char *uncompressed)
 {
 	struct source reader = {
@@ -1507,32 +1558,32 @@ int snappy_uncompress_iov(struct iovec *iov_in, int iov_in_len,
 	};
 	return internal_uncompress(&reader, &output, 0xffffffff);
 }
-EXPORT_SYMBOL(snappy_uncompress_iov);
+EXPORT_SYMBOL(rd_kafka_snappy_uncompress_iov);
 
 /**
- * snappy_uncompress - Uncompress a snappy compressed buffer
+ * rd_kafka_snappy_uncompress - Uncompress a snappy compressed buffer
  * @compressed: Input buffer with compressed data
  * @n: length of compressed buffer
  * @uncompressed: buffer for uncompressed data
  *
  * The uncompressed data buffer must be at least
- * snappy_uncompressed_length(compressed) bytes long.
+ * rd_kafka_snappy_uncompressed_length(compressed) bytes long.
  *
  * Return 0 on success, otherwise an negative error code.
  */
-int snappy_uncompress(const char *compressed, size_t n, char *uncompressed)
+int rd_kafka_snappy_uncompress(const char *compressed, size_t n, char *uncompressed)
 {
 	struct iovec iov = {
 		.iov_base = (char *)compressed,
 		.iov_len = n
 	};
-	return snappy_uncompress_iov(&iov, 1, n, uncompressed);
+	return rd_kafka_snappy_uncompress_iov(&iov, 1, n, uncompressed);
 }
-EXPORT_SYMBOL(snappy_uncompress);
+EXPORT_SYMBOL(rd_kafka_snappy_uncompress);
 
 #else
 /**
- * snappy_compress - Compress a buffer using the snappy compressor.
+ * rd_kafka_snappy_compress - Compress a buffer using the snappy compressor.
  * @env: Preallocated environment
  * @input: Input buffer
  * @input_length: Length of input_buffer
@@ -1542,13 +1593,13 @@ EXPORT_SYMBOL(snappy_uncompress);
  * Return 0 on success, otherwise an negative error code.
  *
  * The output buffer must be at least
- * snappy_max_compressed_length(input_length) bytes long.
+ * rd_kafka_snappy_max_compressed_length(input_length) bytes long.
  *
- * Requires a preallocated environment from snappy_init_env.
+ * Requires a preallocated environment from rd_kafka_snappy_init_env.
  * The environment does not keep state over individual calls
  * of this function, just preallocates the memory.
  */
-int snappy_compress(struct snappy_env *env,
+int rd_kafka_snappy_compress(struct snappy_env *env,
 		    const char *input,
 		    size_t input_length,
 		    char *compressed, size_t *compressed_length)
@@ -1566,20 +1617,20 @@ int snappy_compress(struct snappy_env *env,
 	*compressed_length = (writer.dest - compressed);
 	return err;
 }
-EXPORT_SYMBOL(snappy_compress);
+EXPORT_SYMBOL(rd_kafka_snappy_compress);
 
 /**
- * snappy_uncompress - Uncompress a snappy compressed buffer
+ * rd_kafka_snappy_uncompress - Uncompress a snappy compressed buffer
  * @compressed: Input buffer with compressed data
  * @n: length of compressed buffer
  * @uncompressed: buffer for uncompressed data
  *
  * The uncompressed data buffer must be at least
- * snappy_uncompressed_length(compressed) bytes long.
+ * rd_kafka_snappy_uncompressed_length(compressed) bytes long.
  *
  * Return 0 on success, otherwise an negative error code.
  */
-int snappy_uncompress(const char *compressed, size_t n, char *uncompressed)
+int rd_kafka_snappy_uncompress(const char *compressed, size_t n, char *uncompressed)
 {
 	struct source reader = {
 		.ptr = compressed,
@@ -1591,7 +1642,7 @@ int snappy_uncompress(const char *compressed, size_t n, char *uncompressed)
 	};
 	return internal_uncompress(&reader, &output, 0xffffffff);
 }
-EXPORT_SYMBOL(snappy_uncompress);
+EXPORT_SYMBOL(rd_kafka_snappy_uncompress);
 #endif
 
 static inline void clear_env(struct snappy_env *env)
@@ -1601,7 +1652,7 @@ static inline void clear_env(struct snappy_env *env)
 
 #ifdef SG
 /**
- * snappy_init_env_sg - Allocate snappy compression environment
+ * rd_kafka_snappy_init_env_sg - Allocate snappy compression environment
  * @env: Environment to preallocate
  * @sg: Input environment ever does scather gather
  *
@@ -1610,9 +1661,9 @@ static inline void clear_env(struct snappy_env *env)
  * Returns 0 on success, otherwise negative errno.
  * Must run in process context.
  */
-int snappy_init_env_sg(struct snappy_env *env, bool sg)
+int rd_kafka_snappy_init_env_sg(struct snappy_env *env, bool sg)
 {
-	if (snappy_init_env(env) < 0)
+	if (rd_kafka_snappy_init_env(env) < 0)
 		goto error;
 
 	if (sg) {
@@ -1620,20 +1671,20 @@ int snappy_init_env_sg(struct snappy_env *env, bool sg)
 		if (!env->scratch)
 			goto error;
 		env->scratch_output =
-			vmalloc(snappy_max_compressed_length(kblock_size));
+			vmalloc(rd_kafka_snappy_max_compressed_length(kblock_size));
 		if (!env->scratch_output)
 			goto error;
 	}
 	return 0;
 error:
-	snappy_free_env(env);
+	rd_kafka_snappy_free_env(env);
 	return -ENOMEM;
 }
-EXPORT_SYMBOL(snappy_init_env_sg);
+EXPORT_SYMBOL(rd_kafka_snappy_init_env_sg);
 #endif
 
 /**
- * snappy_init_env - Allocate snappy compression environment
+ * rd_kafka_snappy_init_env - Allocate snappy compression environment
  * @env: Environment to preallocate
  *
  * Passing multiple entries in an iovec is not allowed
@@ -1641,7 +1692,7 @@ EXPORT_SYMBOL(snappy_init_env_sg);
  * Returns 0 on success, otherwise negative errno.
  * Must run in process context.
  */
-int snappy_init_env(struct snappy_env *env)
+int rd_kafka_snappy_init_env(struct snappy_env *env)
 {
     clear_env(env);
 	env->hash_table = vmalloc(sizeof(u16) * kmax_hash_table_size);
@@ -1649,15 +1700,15 @@ int snappy_init_env(struct snappy_env *env)
 		return -ENOMEM;
 	return 0;
 }
-EXPORT_SYMBOL(snappy_init_env);
+EXPORT_SYMBOL(rd_kafka_snappy_init_env);
 
 /**
- * snappy_free_env - Free an snappy compression environment
+ * rd_kafka_snappy_free_env - Free an snappy compression environment
  * @env: Environment to free.
  *
  * Must run in process context.
  */
-void snappy_free_env(struct snappy_env *env)
+void rd_kafka_snappy_free_env(struct snappy_env *env)
 {
 	vfree(env->hash_table);
 #ifdef SG
@@ -1666,4 +1717,4 @@ void snappy_free_env(struct snappy_env *env)
 #endif
 	clear_env(env);
 }
-EXPORT_SYMBOL(snappy_free_env);
+EXPORT_SYMBOL(rd_kafka_snappy_free_env);
