@@ -22,6 +22,7 @@ freely, subject to the following restrictions:
     distribution.
 */
 
+#include "rd.h"
 #include "tinycthread.h"
 #include <stdlib.h>
 
@@ -37,7 +38,6 @@ freely, subject to the following restrictions:
   #include <sys/timeb.h>
 #endif
 
-#include "rd.h"
 
 /* Standard, good-to-have defines */
 #ifndef NULL
@@ -476,6 +476,8 @@ int cnd_timedwait(cnd_t *cond, mtx_t *mtx, const struct timespec *ts)
 
 
 int cnd_timedwait_ms(cnd_t *cnd, mtx_t *mtx, int timeout_ms) {
+  if (timeout_ms == -1 /* INFINITE*/)
+    return cnd_wait(cnd, mtx);
 #if defined(_TTHREAD_WIN32_)
 	return _cnd_timedwait_win32(cnd, mtx, (DWORD)timeout_ms);
 #else
@@ -638,9 +640,13 @@ int thrd_create(thrd_t *thr, thrd_start_t func, void *arg)
 #if defined(_TTHREAD_WIN32_)
   *thr = CreateThread(NULL, 0, _thrd_wrapper_function, (LPVOID) ti, 0, NULL);
 #elif defined(_TTHREAD_POSIX_)
-  if(pthread_create(thr, NULL, _thrd_wrapper_function, (void *)ti) != 0)
   {
-    *thr = 0;
+          int err;
+          if((err = pthread_create(thr, NULL, _thrd_wrapper_function,
+                                   (void *)ti)) != 0) {
+                  errno = err;
+                  *thr = 0;
+          }
   }
 #endif
 
@@ -963,6 +969,52 @@ void call_once(once_flag *flag, void (*func)(void))
 }
 #endif /* defined(_TTHREAD_WIN32_) */
 
+
+#if !defined(_TTHREAD_WIN32_)
+int rwlock_init (rwlock_t *rwl) {
+        int r = pthread_rwlock_init(rwl, NULL);
+        if (r) {
+                errno = r;
+                return thrd_error;
+        }
+        return thrd_success;
+}
+
+int rwlock_destroy (rwlock_t *rwl) {
+        int r = pthread_rwlock_destroy(rwl);
+        if (r) {
+                errno = r;
+                return thrd_error;
+        }
+        return thrd_success;
+}
+
+int rwlock_rdlock (rwlock_t *rwl) {
+        int r = pthread_rwlock_rdlock(rwl);
+        assert(r == 0);
+        return thrd_success;
+}
+
+int rwlock_wrlock (rwlock_t *rwl) {
+        int r = pthread_rwlock_wrlock(rwl);
+        assert(r == 0);
+        return thrd_success;
+}
+
+int rwlock_rdunlock (rwlock_t *rwl) {
+        int r = pthread_rwlock_unlock(rwl);
+        assert(r == 0);
+        return thrd_success;
+}
+
+int rwlock_wrunlock (rwlock_t *rwl) {
+        int r = pthread_rwlock_unlock(rwl);
+        assert(r == 0);
+        return thrd_success;
+}
+
+#endif /* !defined(_TTHREAD_WIN32_) */
+
 #ifdef __cplusplus
 }
 #endif
