@@ -79,12 +79,7 @@ static void produce_messages (uint64_t testid, const char *topic,
                                 errstr, sizeof(errstr));
 
 	/* Create kafka instance */
-	rk = rd_kafka_new(RD_KAFKA_PRODUCER, conf,
-			       errstr, sizeof(errstr));
-	if (!rk)
-		TEST_FAIL("Failed to create rdkafka instance: %s\n", errstr);
-
-	TEST_SAY("Created    kafka instance %s\n", rd_kafka_name(rk));
+	rk = test_create_handle(RD_KAFKA_PRODUCER, conf);
 
 	rkt = rd_kafka_topic_new(rk, topic, topic_conf);
 	if (!rkt)
@@ -335,7 +330,6 @@ static void consume_messages_callback_multi (const char *desc,
 	rd_kafka_topic_t *rkt;
 	rd_kafka_conf_t *conf;
 	rd_kafka_topic_conf_t *topic_conf;
-	char errstr[512];
 	int i;
 
         TEST_SAY("%s: Consume messages %d+%d from %s [%"PRId32"] "
@@ -354,13 +348,7 @@ static void consume_messages_callback_multi (const char *desc,
         }
 
 	/* Create kafka instance */
-	rk = rd_kafka_new(RD_KAFKA_CONSUMER, conf, errstr, sizeof(errstr));
-	if (!rk)
-		TEST_FAIL("%s: Failed to create rdkafka instance: %s\n",
-                          desc, errstr);
-
-	TEST_SAY("%s: Created    kafka instance %s\n",
-                 desc, rd_kafka_name(rk));
+	rk = test_create_handle(RD_KAFKA_CONSUMER, conf);
 
         rd_kafka_topic_conf_set(topic_conf, "auto.offset.reset", "smallest",
                                 NULL, 0);
